__all__ = ()


def _merge_dictionaries(iterator):
    """
    Merges multiple dictionaries into one.
    
    Parameters
    ----------
    iterator : `iterable<None | dict>`
        An iterable over multiple dictionaries.
    
    Returns
    -------
    output : `None | dict`
    """
    output = None
    
    for dictionary in iterator:
        if dictionary is None:
            continue
        
        if output is None:
            output = {}
        
        output.update(dictionary)
    
    return output


def _merge_lists(iterator):
    """
    Merges multiple lists into one.
    
    Parameters
    ----------
    iterator : `iterable<None | list>`
        An iterable over multiple lists.
    
    Returns
    -------
    output : `None | list`
    """
    output = None
    
    for list_ in iterator:
        if list_ is None:
            continue
        
        if output is None:
            output = []
        
        output.extend(list_)
    
    return output



__all__ = ()

from scarletio import include

from ...bases import maybe_snowflake
from ...emoji import Emoji, create_emoji_from_exclusive_inline_data, put_exclusive_emoji_inline_data_into
from ...field_parsers import (
    bool_parser_factory, entity_id_array_parser_factory, entity_id_parser_factory, flag_parser_factory,
    force_string_parser_factory, int_parser_factory, int_postprocess_parser_factory,
    nullable_entity_array_parser_factory, nullable_string_parser_factory, nullable_unix_time_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, flag_optional_putter_factory,
    force_string_putter_factory, int_optional_postprocess_putter_factory, int_putter_factory,
    nullable_entity_array_optional_putter_factory, nullable_string_putter_factory,
    nullable_unix_time_optional_putter_factory, nulled_int_optional_putter_factory,
    optional_entity_id_array_optional_putter_factory, preinstanced_optional_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_array_validator_factory, entity_id_validator_factory, flag_validator_factory,
    force_string_validator_factory, int_conditional_validator_factory, int_options_validator_factory,
    nullable_date_time_validator_factory, nullable_entity_array_validator_factory, nullable_entity_validator_factory,
    nullable_string_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase, User, create_partial_user_from_id
from ...utils import datetime_to_timestamp, timestamp_to_datetime

from ..forum_tag import ForumTag
from ..permission_overwrite import PermissionOverwrite

from .constants import (
    AUTO_ARCHIVE_DEFAULT, AUTO_ARCHIVE_OPTIONS, BITRATE_DEFAULT, BITRATE_MAX, BITRATE_MIN, NAME_LENGTH_MAX,
    NAME_LENGTH_MIN, SLOWMODE_DEFAULT, SLOWMODE_MAX, SLOWMODE_MIN, STATUS_LENGTH_MAX, STATUS_LENGTH_MIN,
    TOPIC_LENGTH_MAX, TOPIC_LENGTH_MIN, USER_LIMIT_DEFAULT, USER_LIMIT_MAX, USER_LIMIT_MIN
)
from .flags import ChannelFlag
from .preinstanced import ForumLayout, SortOrder, VideoQualityMode, VoiceRegion


Channel = include('Channel')

# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id = entity_id_optional_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', NotImplemented, include = 'Application')

# applied_tag_ids

parse_applied_tag_ids = entity_id_array_parser_factory('applied_tags')
put_applied_tag_ids = optional_entity_id_array_optional_putter_factory('applied_tags')
validate_applied_tag_ids = entity_id_array_validator_factory('applied_tag_ids', ForumTag)

# archived

def parse_archived(data):
    """
    Parses out the `archived` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    archived : `bool`
    """
    sub_data = data.get('thread_metadata', None)
    if sub_data is None:
        archived = False
    else:
        archived = sub_data.get('archived', False)
    
    return archived


def put_archived(archived, data, defaults):
    """
    Puts the `archived`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    archived : `bool`
        Whether the channel is archived.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if archived or defaults:
        try:
            sub_data = data['thread_metadata']
        except KeyError:
            sub_data = {}
            data['thread_metadata'] = sub_data
        
        sub_data['archived'] = archived
    
    return data


validate_archived = bool_validator_factory('archived', False)

# archived_at

def parse_archived_at(data):
    """
    Parses out the `archived_at` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    archived_at : `int`
    """
    sub_data = data.get('thread_metadata', None)
    if sub_data is None:
        archived_at = None
    else:
        archived_at_timestamp = sub_data.get('archive_timestamp', None)
        if (archived_at_timestamp is None):
            archived_at = None
        else:
            archived_at = timestamp_to_datetime(archived_at_timestamp)
    
    return archived_at


def put_archived_at(archived_at, data, defaults):
    """
    Puts the `archived_at`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    archived_at : `None | DateTime`
        When the channel was archived.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (archived_at is not None):
        try:
            sub_data = data['thread_metadata']
        except KeyError:
            sub_data = {}
            data['thread_metadata'] = sub_data
        
        if archived_at is None:
            archived_at_timestamp = None
        else:
            archived_at_timestamp = datetime_to_timestamp(archived_at)
            
        sub_data['archive_timestamp'] = archived_at_timestamp
    
    return data


validate_archived_at = nullable_date_time_validator_factory('archived_at')

# auto_archive_after

def parse_auto_archive_after(data):
    """
    Parses out the `auto_archive_after` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    auto_archive_after : `int`
    """
    sub_data = data.get('thread_metadata', None)
    if sub_data is None:
        auto_archive_after = AUTO_ARCHIVE_DEFAULT
    else:
        auto_archive_after = sub_data.get('auto_archive_duration', None)
        if auto_archive_after is None:
            auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            auto_archive_after *= 60

    return auto_archive_after


def put_auto_archive_after(auto_archive_after, data, defaults, *, flatten_thread_metadata = False):
    """
    Puts the `auto_archive_after`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    flatten_thread_metadata : `bool` = `False`, Optional (Keyword only)
        Whether the field should be flattened instead of nested.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (auto_archive_after != AUTO_ARCHIVE_DEFAULT):
        if flatten_thread_metadata:
            data_to_use = data
        
        else:
            try:
                data_to_use = data['thread_metadata']
            except KeyError:
                data_to_use = {}
                data['thread_metadata'] = data_to_use
        
        data_to_use['auto_archive_duration'] = auto_archive_after // 60
    
    return data


validate_auto_archive_after = int_options_validator_factory('auto_archive_after', AUTO_ARCHIVE_OPTIONS, 0)

# available_tags

parse_available_tags = nullable_entity_array_parser_factory('available_tags', ForumTag)
put_available_tags = nullable_entity_array_optional_putter_factory('available_tags', ForumTag)
validate_available_tags = nullable_entity_array_validator_factory('available_tags', ForumTag)

# bitrate

parse_bitrate = int_parser_factory('bitrate', BITRATE_DEFAULT)
put_bitrate = int_putter_factory('bitrate')
validate_bitrate = int_conditional_validator_factory(
    'bitrate',
    BITRATE_MIN,
    (lambda bitrate: bitrate >= BITRATE_MIN and bitrate <= BITRATE_MAX),
    f'>= {BITRATE_MIN} and <= {BITRATE_MAX},'
)

# created_at

def parse_created_at(data):
    """
    Parses out the `created_at` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    created_at : `None | DateTime`
    """
    sub_data = data.get('thread_metadata', None)
    if sub_data is None:
        created_at = None
    else:
        created_at_timestamp = sub_data.get('create_timestamp', None)
        if (created_at_timestamp is None):
            created_at = None
        else:
            created_at = timestamp_to_datetime(created_at_timestamp)
    
    return created_at


def put_created_at(created_at, data, defaults):
    """
    Puts the `created_at`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    created_at : `None | DateTime`
        When the channel was created.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (created_at is not None):
        try:
            sub_data = data['thread_metadata']
        except KeyError:
            sub_data = {}
            data['thread_metadata'] = sub_data
        
        if created_at is None:
            created_at_timestamp = None
        else:
            created_at_timestamp = datetime_to_timestamp(created_at)
            
        sub_data['create_timestamp'] = created_at_timestamp
    
    return data


validate_created_at = nullable_date_time_validator_factory('created_at')

# default_forum_layout

parse_default_forum_layout = preinstanced_parser_factory('default_forum_layout', ForumLayout, ForumLayout.none)
put_default_forum_layout = preinstanced_optional_putter_factory('default_forum_layout', ForumLayout.none)
validate_default_forum_layout = preinstanced_validator_factory('default_forum_layout', ForumLayout)

# default_sort_order

parse_default_sort_order = preinstanced_parser_factory('default_sort_order', SortOrder, SortOrder.latest_activity)
put_default_sort_order = preinstanced_optional_putter_factory('default_sort_order', SortOrder.latest_activity)
validate_default_sort_order = preinstanced_validator_factory('default_sort_order', SortOrder)

# default_thread_auto_archive_after

parse_default_thread_auto_archive_after = int_postprocess_parser_factory(
    'default_auto_archive_duration',
    AUTO_ARCHIVE_DEFAULT,
    (lambda default_thread_auto_archive_after: default_thread_auto_archive_after * 60),
)
put_default_thread_auto_archive_after = int_optional_postprocess_putter_factory(
    'default_auto_archive_duration',
    AUTO_ARCHIVE_DEFAULT,
    (lambda default_thread_auto_archive_after: default_thread_auto_archive_after // 60),
)
validate_default_thread_auto_archive_after = int_options_validator_factory(
    'default_thread_auto_archive_after', AUTO_ARCHIVE_OPTIONS, 0
)

# default_thread_reaction_emoji

def parse_default_thread_reaction_emoji(data):
    """
    Parses out the `default_thread_reaction_emoji` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    default_thread_reaction_emoji : ``None | Emoji``
    """
    default_thread_reaction_emoji_data = data.get('default_reaction_emoji', None)
    if (default_thread_reaction_emoji_data is None):
        default_thread_reaction_emoji = None
    else:
        default_thread_reaction_emoji = create_emoji_from_exclusive_inline_data(default_thread_reaction_emoji_data)
    
    return default_thread_reaction_emoji


def put_default_thread_reaction_emoji(default_thread_reaction_emoji, data, defaults):
    """
    Puts the `default_thread_reaction_emoji`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    default_thread_reaction_emoji : ``None | Emoji``
        The emoji to show in the add reaction button on a thread of the forum channel.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (default_thread_reaction_emoji is not None):
        if default_thread_reaction_emoji is None:
            emoji_data = None
        else:
            emoji_data = put_exclusive_emoji_inline_data_into(default_thread_reaction_emoji, {})
        
        data['default_reaction_emoji'] = emoji_data
    
    return data


validate_default_thread_reaction_emoji = nullable_entity_validator_factory('default_thread_reaction_emoji', Emoji)

# default_thread_slowmode

parse_default_thread_slowmode = int_parser_factory('default_thread_rate_limit_per_user', SLOWMODE_DEFAULT)
put_default_thread_slowmode = nulled_int_optional_putter_factory(
    'default_thread_rate_limit_per_user',
    SLOWMODE_DEFAULT,
)
validate_default_thread_slowmode = int_conditional_validator_factory(
    'default_thread_slowmode',
    SLOWMODE_MIN,
    (
        lambda default_thread_slowmode:
        default_thread_slowmode >= SLOWMODE_MIN and default_thread_slowmode <= SLOWMODE_MAX
    ),
    f'>= {SLOWMODE_MIN} and <= {SLOWMODE_MAX},'
)

# flags

parse_flags = flag_parser_factory('flags', ChannelFlag)
put_flags = flag_optional_putter_factory('flags', ChannelFlag())
validate_flags = flag_validator_factory('flags', ChannelFlag)

# invitable

def parse_invitable(data):
    """
    Parses out the `invitable` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    invitable : `bool`
    """
    sub_data = data.get('thread_metadata', None)
    if sub_data is None:
        invitable = True
    else:
        invitable = sub_data.get('invitable', True)
    
    return invitable


def put_invitable(invitable, data, defaults, *, flatten_thread_metadata = False):
    """
    Puts the `invitable`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    invitable : `bool`
        Whether non-moderators can invite other non-moderators to the threads.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    flatten_thread_metadata : `bool` = `False`, Optional (Keyword only)
        Whether the field should be flattened instead of nested.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (not invitable) or defaults:
        if flatten_thread_metadata:
            data_to_use = data
        
        else:
            try:
                data_to_use = data['thread_metadata']
            except KeyError:
                data_to_use = {}
                data['thread_metadata'] = data_to_use
        
        data_to_use['invitable'] = invitable
    
    return data


validate_invitable = bool_validator_factory('invitable', True)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# nsfw

parse_nsfw = bool_parser_factory('nsfw', False)
put_nsfw = bool_optional_putter_factory('nsfw', False)
validate_nsfw = bool_validator_factory('nsfw', False)

# open

def parse_open(data):
    """
    Parses out the `open` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
         Channel data.
    
    Returns
    -------
    open_ : `bool`
    """
    sub_data = data.get('thread_metadata', None)
    if sub_data is None:
        open_ = True
    else:
        open_ = not sub_data.get('locked', False)
    
    return open_


def put_open(open_, data, defaults, *, flatten_thread_metadata = False):
    """
    Puts the `open`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    open_ : `bool`
        Whether the channel is open.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    flatten_thread_metadata : `bool` = `False`, Optional (Keyword only)
        Whether the field should be flattened instead of nested.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (not open_) or defaults:
        if flatten_thread_metadata:
            data_to_use = data
        
        else:
            try:
                data_to_use = data['thread_metadata']
            except KeyError:
                data_to_use = {}
                data['thread_metadata'] = data_to_use
        
        data_to_use['locked'] = not open_
    
    return data


validate_open = bool_validator_factory('open', True)

# owner_id

parse_owner_id = entity_id_parser_factory('owner_id')
put_owner_id = entity_id_optional_putter_factory('owner_id')
validate_owner_id = entity_id_validator_factory('owner_id', ClientUserBase)

# parent_id

parse_parent_id = entity_id_parser_factory('parent_id')
put_parent_id = entity_id_optional_putter_factory('parent_id')
validate_parent_id = entity_id_validator_factory('parent_id', NotImplemented, include = 'Channel')

# permission_overwrites

def parse_permission_overwrites(data):
    """
    Parses the permission overwrites from the given data and returns them.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    permission_overwrites : `None`, `dict` of (`int`, ``PermissionOverwrite``) items
    """
    permission_overwrites = None
    
    permission_overwrites_datas = data.get('permission_overwrites', None)
    if (permission_overwrites_datas is not None) and permission_overwrites_datas:
        for permission_overwrite_data in permission_overwrites_datas:
            permission_overwrite = PermissionOverwrite.from_data(permission_overwrite_data)
            if permission_overwrites is None:
                permission_overwrites = {}
            permission_overwrites[permission_overwrite.target_id] = permission_overwrite
    
    return permission_overwrites


def put_permission_overwrites(permission_overwrites, data, defaults):
    """
    Puts the `permission_overwrites`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    permission_overwrites :`dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if permission_overwrites is None:
        permission_overwrite_datas = []
    else:
        permission_overwrite_datas = [
            permission_overwrite.to_data(include_internals = True)
            for permission_overwrite in permission_overwrites.values()
        ]
    
    data['permission_overwrites'] = permission_overwrite_datas
    return data


def validate_permission_overwrites(permission_overwrites):
    """
    Validates the given `permission_overwrites` field.
    
    Parameters
    ----------
    permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``
        The permission_overwrites to validate.
    
    Returns
    -------
    permission_overwrites : ``None | dict<int, PermissionOverwrite>``
    """
    if permission_overwrites is None:
        return None
    
    if (getattr(permission_overwrites, '__iter__', None) is None):
        raise TypeError(
            f'`permission_overwrites` can be `None`, `iterable` of `{PermissionOverwrite.__name__}`, got '
            f'{permission_overwrites.__class__.__name__}; {permission_overwrites!r}.'
        )
    
    permission_overwrites_processed = None
    
    for permission_overwrite in permission_overwrites:
        if not isinstance(permission_overwrite, PermissionOverwrite):
            raise TypeError(
                f'`permission_overwrites` can contain `{PermissionOverwrite.__name__}` elements, got '
                f'{permission_overwrite.__class__.__name__}; {permission_overwrite!r}; '
                f'permission_overwrites = {permission_overwrites!r}.'
            )
        
        if permission_overwrites_processed is None:
            permission_overwrites_processed = {}
        
        permission_overwrites_processed[permission_overwrite.target_id] = permission_overwrite
    
    return permission_overwrites_processed

# position

parse_position = int_parser_factory('position', 0)
put_position = int_putter_factory('position')
validate_position = int_conditional_validator_factory(
    'position',
    0,
    lambda position : position >= 0,
    '>= 0',
)

# region

parse_region = preinstanced_parser_factory('rtc_region', VoiceRegion, VoiceRegion.unknown)
put_region = preinstanced_optional_putter_factory('rtc_region', VoiceRegion.unknown)
validate_region = preinstanced_validator_factory('region', VoiceRegion)

# slowmode

parse_slowmode = int_parser_factory('rate_limit_per_user', SLOWMODE_DEFAULT)
put_slowmode = nulled_int_optional_putter_factory(
    'rate_limit_per_user',
    SLOWMODE_DEFAULT,
)
validate_slowmode = int_conditional_validator_factory(
    'slowmode',
    SLOWMODE_MIN,
    (
        lambda slowmode:
        slowmode >= SLOWMODE_MIN and slowmode <= SLOWMODE_MAX
    ),
    f'>= {SLOWMODE_MIN} and <= {SLOWMODE_MAX}',
)


# voice_engaged_since

parse_voice_engaged_since = nullable_unix_time_parser_factory('voice_start_time')
put_voice_engaged_since = nullable_unix_time_optional_putter_factory('voice_start_time')
validate_voice_engaged_since = nullable_date_time_validator_factory('voice_engaged_since')


# status

parse_status = nullable_string_parser_factory('status')
put_status = url_optional_putter_factory('status')
validate_status = nullable_string_validator_factory('status', STATUS_LENGTH_MIN, STATUS_LENGTH_MAX)

# topic

parse_topic = nullable_string_parser_factory('topic')
put_topic = nullable_string_putter_factory('topic')
validate_topic = nullable_string_validator_factory('topic', TOPIC_LENGTH_MIN, TOPIC_LENGTH_MAX)

# user_limit

parse_user_limit = int_parser_factory('user_limit', USER_LIMIT_DEFAULT)
put_user_limit = int_putter_factory('user_limit')
validate_user_limit = int_conditional_validator_factory(
    'user_limit',
    USER_LIMIT_MIN,
    (lambda user_limit: user_limit >= USER_LIMIT_MIN and user_limit <= USER_LIMIT_MAX),
    f'>= {USER_LIMIT_MIN} and <= {USER_LIMIT_MAX},'
)

# users

def parse_users(data):
    """
    Parses out the `users` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    users : ``list<ClientUserBase>``
    """
    users = []
    
    try:
        user_datas = data['recipients']
    except KeyError:
        pass
    else:
        for user_data in user_datas:
            user = User.from_data(user_data)
            users.append(user)
        
        users.sort()
    
    return users


def put_users(users, data, defaults):
    """
    Puts the `users`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    users : ``list<ClientUserBase>``
        The users in the channel.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['recipients'] = [user.to_data(defaults = defaults, include_internals = True) for user in users]
    
    return data


def validate_users(users):
    """
    Validates the given `users` field.
    
    Parameters
    ----------
    users : `iterable` of (``ClientUserBase`` | `int`)
        The users in the channel.
    
    Returns
    -------
    users : ``list<ClientUserBase>``
    
    Raises
    ------
    TypeError
        - If `users` is not `list` of (``ClientUserBase``, `int`).
    """
    if users is None:
        return []
    
    if (getattr(users, '__iter__', None) is None):
        raise TypeError(
            f'`users` can be `None`, `iterable` of (`int` | `{ClientUserBase.__name__}`), '
            f'got {type(users).__name__}; {users!r}.'
        )
    
    users_processed = set()
    
    for user in users:
        if not isinstance(user, ClientUserBase):
            user_id = maybe_snowflake(user)
            if user_id is None:
                raise TypeError(
                    f'`users` can contain `int`, `{ClientUserBase.__name__}` elements, got '
                    f'{type(user).__name__}; {user!r}; users = {users!r}.'
                )
            
            user = create_partial_user_from_id(user_id)
        
        users_processed.add(user)
    
    return sorted(users_processed)


# video_quality_mode


VIDEO_QUALITY_MODE_NONE = VideoQualityMode.none
VIDEO_QUALITY_MODE_AUTO = VideoQualityMode.auto

parse_video_quality_mode = preinstanced_parser_factory('video_quality_mode', VideoQualityMode, VideoQualityMode.auto)


def put_video_quality_mode(video_quality_mode, data, defaults):
    """
    Puts the `video_quality_mode`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    video_quality_mode : ``VideoQualityMode``
        The video quality of the voice channel.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (
        (video_quality_mode is not VIDEO_QUALITY_MODE_NONE) and
        (video_quality_mode is not VIDEO_QUALITY_MODE_AUTO)
    ):
        if video_quality_mode is VIDEO_QUALITY_MODE_NONE:
            video_quality_mode = VIDEO_QUALITY_MODE_AUTO
        
        data['video_quality_mode'] = video_quality_mode.value
    
    return data


validate_video_quality_mode = preinstanced_validator_factory('video_quality_mode', VideoQualityMode)



__all__ = ('SoundboardSound',)

from ...bases import DiscordEntity
from ...core import GUILDS, SOUNDBOARD_SOUNDS
from ...http.urls import build_soundboard_sound_url
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ZEROUSER, create_partial_user_from_id

from .constants import DEFAULT_SOUNDBOARD_SOUND_LIMIT
from .fields import (
    parse_available, parse_emoji, parse_guild_id, parse_id, parse_name, parse_user, parse_user_id, parse_volume,
    put_available, put_emoji, put_guild_id, put_id, put_name, put_user_id, put_user,
    put_volume, validate_available, validate_emoji, validate_guild_id, validate_id, validate_name,
    validate_user_id, validate_volume
)


PRECREATE_FIELDS = {
    'available': ('available', validate_available),
    'emoji': ('emoji', validate_emoji),
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'name': ('name', validate_name),
    'user': ('user_id', validate_user_id),
    'user_id': ('user_id', validate_user_id),
    'volume': ('volume', validate_volume),
}


class SoundboardSound(DiscordEntity, immortal = True):
    """
    Represents a sound board sound.
    
    Attributes
    ----------
    _cache_user : ``None | ClientUserBase``
        Cache field used by the ``.user`` property.
    available : `bool`
        Whether the sound is available.
    emoji : ``None | Emoji``
        Emoji assigned to the sound.
    guild_id : `int`
        The guild's identifier to which the sound is added to. At the case of builtin sounds this is `0`.
    name : `str`
        The name of the sound.
    user_id : `int`
        The user's identifier who created the sound.
    volume : `float`
        The volume of the sound to play as.
    """
    __slots__ = ('_cache_user', 'available', 'emoji', 'guild_id', 'name', 'user_id', 'volume')
    
    def __new__(cls, *, available = ..., emoji = ..., name = ..., user_id = ..., volume = ...):
        """
        Creates a new sound board sound with the given fields.
        
        Parameters
        ----------
        available : `bool`, Optional (Keyword only)
            Whether the sound is available.
        emoji : ``None | Emoji``, Optional (Keyword only)
            Emoji assigned to the sound.
        name : `str`, Optional (Keyword only)
            The name of the sound.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user or their identifier who created the sound.
        volume : `float`, Optional (Keyword only)
            The volume of the sound to play as.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # available
        if available is ...:
            available = True
        else:
            available = validate_available(available)
        
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # user
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # volume
        if volume is ...:
            volume = 1.0
        else:
            volume = validate_volume(volume)
        
        # Construct
        self = object.__new__(cls)
        self._cache_user = None
        self.available = available
        self.emoji = emoji
        self.guild_id = 0
        self.id = 0
        self.name = name
        self.user_id = user_id
        self.volume = volume
        return self
    
    
    @classmethod
    def from_data(cls, data, *, strong_cache = True):
        """
        Creates a new soundboard sound from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sound data.
        
        strong_cache : `bool` = `True`, Optional (Keyword only)
            Whether the instance should be put into its strong cache.
        
        Returns
        -------
        self : `instance<cls>`
        """
        sound_id = parse_id(data)
        
        try:
            self = SOUNDBOARD_SOUNDS[sound_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sound_id
            self._set_attributes(data)
            SOUNDBOARD_SOUNDS[sound_id] = self
        else:
            if strong_cache and (not self.partial):
                return self
            
            self._set_attributes(data)
        
        if strong_cache:
            try:
                guild = GUILDS[self.guild_id]
            except KeyError:
                pass
            else:
                soundboard_sounds = guild.soundboard_sounds
                if soundboard_sounds is None:
                    soundboard_sounds = {}
                    guild.soundboard_sounds = soundboard_sounds
                
                soundboard_sounds[sound_id] = self
            
        return self
    
    
    @classmethod
    def from_data_is_created(cls, data):
        """
        Creates a new soundboard sound from the given data.
        
        Also returns whether the instance was new (or partial) or already existed.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sound data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        sound_id = parse_id(data)
        
        try:
            self = SOUNDBOARD_SOUNDS[sound_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sound_id
            self._set_attributes(data)
            SOUNDBOARD_SOUNDS[sound_id] = self
        
        else:
            if not self.partial:
                return self, False
            
            self._set_attributes(data)
        
        try:
            guild = GUILDS[self.guild_id]
        except KeyError:
            pass
        else:
            soundboard_sounds = guild.soundboard_sounds
            if soundboard_sounds is None:
                soundboard_sounds = {}
                guild.soundboard_sounds = soundboard_sounds
            
            soundboard_sounds[sound_id] = self
        
        return self, True
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the soundboard sound to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with the default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields, like id-s should be present as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_emoji(self.emoji, data, defaults)
        put_name(self.name, data, defaults)
        put_volume(self.volume, data, defaults)
        
        if include_internals:
            put_available(self.available, data, defaults)
            put_guild_id(self.guild_id, data, defaults)
            put_id(self.id, data, defaults)
            put_user(self.user, data, defaults)
            put_user_id(self.user_id, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data):
        """
        Sets the soundboard's attributes excluding ``.id`` from from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sound data.
        """
        self._cache_user = parse_user(data)
        self.guild_id = parse_guild_id(data)
        self.user_id = parse_user_id(data)
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the soundboard sound's modifiable attributes only.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sound data.
        """
        self.available = parse_available(data)
        self.emoji = parse_emoji(data)
        self.name = parse_name(data)
        self.volume = parse_volume(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the soundboard sound's modifiable attributes only. Returns the changed attributes in a
        `attribute-name` - `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sound data.
        
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-----------+-------------------+
        | Keys      | Values            |
        +===========+===================+
        | available | `bool`            |
        +-----------+-------------------+
        | emoji     | ``None | Emoji``  |
        +-----------+-------------------+
        | name      | `str`             |
        +-----------+-------------------+
        | volume    | `float`           |
        +-----------+-------------------+
        """
        old_attributes = {}
        
        # available
        available = parse_available(data)
        if (available != self.available):
            old_attributes['available'] = self.available
            self.available = available
        
        # emoji
        emoji = parse_emoji(data)
        if (emoji != self.emoji):
            old_attributes['emoji'] = self.emoji
            self.emoji = emoji
        
        # name
        name = parse_name(data)
        if (name != self.name):
            old_attributes['name'] = self.name
            self.name = name
        
        # volume
        volume = parse_volume(data)
        if (volume != self.volume):
            old_attributes['volume'] = self.volume
            self.volume = volume
        
        return old_attributes
    
    
    def __repr__(self):
        """Returns the soundboard sound's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' id = ')
        repr_parts.append(repr(self.id))
        
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two soundboard sounds are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two soundboard sounds are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Compares the soundboard sound's attributes to the other one's.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other sound to compare self to.
            
            > Must have the same type as self.
        
        Returns
        -------
        is_equal : `bool`
        """
        # available
        if self.available != other.available:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # guild_id -> skip
        # id -> skip
        
        # name
        if self.name != other.name:
            return False
        
        # user -> skip
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        # volume
        if self.volume != other.volume:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the soundboard sound's hash value."""
        team_id = self.id
        if team_id:
            return team_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Calculates the sound board sound's hash based on their fields.
        
        This method is called by ``.__hash__`` if the object has no ``.id`` set.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # available
        hash_value ^= self.available << 4
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        # guild_id -> skip
        # id -> skip
        
        # name
        hash_value ^= hash(self.name)
        
        # user -> skip
        
        # user
        hash_value ^= self.user_id
        
        # volume
        hash_value ^= hash(self.volume)
        
        return hash_value
    
    
    @classmethod
    def _create_empty(cls, sound_id, guild_id = 0):
        """
        Crates a empty soundboard sound with its attributes set to their default values.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._cache_user = None
        self.available = True
        self.emoji = None
        self.guild_id = guild_id
        self.id = sound_id
        self.name = ''
        self.user_id = 0
        self.volume = 1.0
        return self
    
    
    @classmethod
    def precreate(cls, sound_id, **keyword_parameters):
        """
        Creates a soundboard sound instance. If already exists, returns that.
        
        Parameters
        ----------
        sound_id : `int`
            The soundboard sound's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        available : `bool`, Optional (Keyword only)
            Whether the sound is available.
        emoji : ``None | Emoji``, Optional (Keyword only)
            Emoji assigned to the sound.
        guild : `int`, `None`, ``Guild``, Optional (Keyword only)
            Alternative of `guild_id`.
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            The guild or its identifier to which the sound is added to.
        name : `str`, Optional (Keyword only)
            The name of the sound.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            Alternative of `user`.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user or their identifier who created the sound.
        volume : `float`, Optional (Keyword only)
            The volume of the sound to play as.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        sound_id = validate_id(sound_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = SOUNDBOARD_SOUNDS[sound_id]
        except KeyError:
            self = cls._create_empty(sound_id)
            SOUNDBOARD_SOUNDS[sound_id] = self
        else:
            if (not self.partial):
                return self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def copy(self):
        """
        Copies the soundboard sound.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._cache_user = self._cache_user
        new.available = self.available
        new.emoji = self.emoji
        new.guild_id = 0
        new.id = 0
        new.name = self.name
        new.user_id = self.user_id
        new.volume = self.volume
        return new
    
    
    def copy_with(self, *, available = ..., emoji = ..., name = ..., user_id = ..., volume = ...):
        """
        Copies the soundboard sound with the given parameters.
        
        Parameters
        ----------
        available : `bool`, Optional (Keyword only)
            Whether the sound is available.
        emoji : ``None | Emoji``, Optional (Keyword only)
            Emoji assigned to the sound.
        name : `str`, Optional (Keyword only)
            The name of the sound.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user or their identifier who created the sound.
        volume : `float`, Optional (Keyword only)
            The volume of the sound to play as.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # available
        if available is ...:
            available = self.available
        else:
            available = validate_available(available)
        
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # user
        if user_id is ...:
            user_id = self.user_id
            user = self._cache_user
        else:
            user_id = validate_user_id(user_id)
            user = None
        
        # volume
        if volume is ...:
            volume = self.volume
        else:
            volume = validate_volume(volume)
        
        # Construct
        new = object.__new__(type(self))
        new._cache_user = user
        new.available = available
        new.emoji = emoji
        new.guild_id = 0
        new.id = 0
        new.name = name
        new.user_id = user_id
        new.volume = volume
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the sound is partial.
        
        Returns
        -------
        partial : `bool
        """
        guild_id = self.guild_id
        if not guild_id:
            return True
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return True
        
        soundboard_sounds = guild.soundboard_sounds
        if soundboard_sounds is None:
            return True
        
        if self.id not in soundboard_sounds:
            return True
        
        return guild.partial
    
    
    def _delete(self):
        """
        Removes the soundboard sound's references.
        """
        try:
            guild = GUILDS[self.guild_id]
        except KeyError:
            pass
        else:
            soundboard_sounds = guild.soundboard_sounds
            if (soundboard_sounds is not None):
                try:
                    del soundboard_sounds[self.id]
                except KeyError:
                    pass
                else:
                    if not soundboard_sounds:
                        guild.soundboard_sounds = None
    
    
    @property
    def guild(self):
        """
        Returns the guild to which the sound is added to.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the user who created the sound.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        user = self._cache_user
        if (user is not None):
            return user
        
        user_id = self.user_id
        if not user_id:
            return ZEROUSER
        
        user = create_partial_user_from_id(user_id)
        self._cache_user = user
        return user

    
    def is_custom_sound(self):
        """
        Returns whether the soundboard sound is a custom sound.
        
        Returns
        -------
        is_custom_sound : `bool`
        """
        sound_id = self.id
        return sound_id == 0 or sound_id >= DEFAULT_SOUNDBOARD_SOUND_LIMIT
    
    
    def is_default_sound(self):
        """
        Returns whether the soundboard sound is a default sound.
        
        Returns
        -------
        is_default_sound : `bool`
        """
        sound_id = self.id
        return sound_id != 0 and sound_id < DEFAULT_SOUNDBOARD_SOUND_LIMIT
    
    
    @property
    def mention(self):
        """
        Returns the mention of the soundboard sound.
        
        Returns
        -------
        mention : `str`
        """
        return f'<sound:{self.guild_id!s}:{self.id!s}>'
    
    
    @property
    def url(self):
        """
        Returns the soundboard sound's url.
        
        Returns
        -------
        url : `str`
        """
        return build_soundboard_sound_url(self.id)



__all__ = ('ZEROUSER', 'create_partial_user_from_id', )

from functools import partial as partial_func

from scarletio import export

from ...core import USERS

from .fields import (
    put_avatar_decoration, put_banner_color, put_display_name, put_name,
    validate_avatar_decoration, validate_banner_color, validate_display_name, validate_name
)
from .orin_user_base import USER_BANNER
from .user import User
from .user_base import USER_AVATAR


@export
def create_partial_user_from_id(user_id):
    """
    Creates a partial user from the given `user_id`. If the user already exists returns that instead.
    
    Parameters
    ----------
    user_id : `int`
        The unique identifier number of the user.
    
    Returns
    -------
    user : ``ClientUserBase``
    """
    try:
        return USERS[user_id]
    except KeyError:
        pass
    
    user = User._create_empty(user_id)
    USERS[user_id] = user
    return user


ZEROUSER = create_partial_user_from_id(0)
export(ZEROUSER, 'ZEROUSER')


USER_SELF_FIELD_CONVERTERS = {
    'avatar': (
        partial_func(USER_AVATAR.validate_icon, allow_data = True),
        partial_func(USER_AVATAR.put_into, as_data = True),
    ),
    'avatar_decoration': (validate_avatar_decoration, put_avatar_decoration),
    'banner': (
        partial_func(USER_BANNER.validate_icon, allow_data = True),
        partial_func(USER_BANNER.put_into, as_data = True),
    ),
    'banner_color': (validate_banner_color, put_banner_color),
    'display_name': (validate_display_name, put_display_name),
    'name': (validate_name, put_name),
}



__all__ = ('GuildActivityOverviewActivityLevel',)

from ...bases import Preinstance as P, PreinstancedBase


class GuildActivityOverviewActivityLevel(PreinstancedBase, value_type = int):
    """
    Represents a guild activity overview's activity's level.
    
    Attributes
    ----------
    name : `str`
        The name of the guild activity overview activity's level.
    
    value : `int`
        The identifier value the guild activity overview activity's level.
    
    Type Attributes
    ---------------
    Every predefined guild overview activity level can be accessed as type attribute as well:
    
    +-----------------------+-------------------+-------------------+
    | Type attribute name   | Name              | Value             |
    +=======================+===================+===================+
    | none                  | none              | 0                 |
    +-----------------------+-------------------+-------------------+
    | any_previous          | any previous      | 1                 |
    +-----------------------+-------------------+-------------------+
    | recently_popular      | recently popular  | 2                 |
    +-----------------------+-------------------+-------------------+
    """
    __slots__ = ()
    
    none = P(0, 'none')
    any_previous = P(1, 'any previous')
    recently_popular = P(2, 'recently popular')



__all__ = ()

from ...field_validators import nullable_entity_validator_factory
from ...poll import Poll

from ..poll_update import PollUpdate


# added

validate_added = nullable_entity_validator_factory('added', Poll)


# updated

validate_updated = nullable_entity_validator_factory('updated', PollUpdate)


# removed

validate_removed = nullable_entity_validator_factory('removed', Poll)



__all__ = ()

import sys
from os import listdir as list_directory
from os.path import join as join_paths, isdir as is_directory, exists

from ..plugin_loader.constants import IGNORED_DIRECTORY_NAMES, PLUGIN_ROOTS


def _iter_sub_directories(directories):
    """
    Iterates over the sub directories of the given ones.
    
    Parameters
    ----------
    directories : `iterable` of `str`
        Directories to look sub directories for in. These directories are also yielded.
    
    Yields
    ------
    directory : `str`
    """
    directories = [*directories]
    
    while directories:
        directory = directories.pop()
        yield directory
        
        for name in list_directory(directory):
            if name in IGNORED_DIRECTORY_NAMES:
                continue
            
            path = join_paths(directory, name)
            if not is_directory(path):
                continue
            
            directories.append(path)
            continue


def _iter_plugin_root_paths():
    """
    Iterates over the plugin root paths.
    
    This function is an iterable generator.
    
    Yields
    ------
    directory : `str`
    """
    for plugin_root in PLUGIN_ROOTS:
        for path_base in sys.path:
            path = join_paths(path_base, *plugin_root)
            if exists(path):
                yield path
                break



__all__ = ()

from ...application import Application
from ...channel import (
    Channel, ChannelType, create_partial_channel_data, create_partial_channel_from_data, create_partial_channel_from_id
)
from ...field_parsers import (
    bool_parser_factory, default_entity_parser_factory, flag_parser_factory, force_string_parser_factory,
    int_parser_factory, nullable_date_time_parser_factory, nullable_entity_parser_factory,
    nullable_functional_parser_factory, nullable_int_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, default_entity_putter_factory, entity_id_optional_putter_factory,
    entity_id_putter_factory, flag_optional_putter_factory, force_string_putter_factory, int_putter_factory,
    nullable_date_time_optional_putter_factory, nullable_entity_optional_putter_factory,
    nullable_field_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator_factory, entity_id_validator_factory, flag_validator_factory,
    force_string_validator_factory, int_conditional_validator_factory, nullable_date_time_validator_factory,
    nullable_entity_validator_factory, preinstanced_validator_factory
)
from ...guild import (
    Guild, GuildActivityOverview, create_partial_guild_data, create_partial_guild_from_data, create_partial_guild_from_id
)
from ...permission import Permission
from ...user import ClientUserBase, User, ZEROUSER

from .flags import InviteFlag
from .preinstanced import InviteTargetType, InviteType


# approximate_online_count

parse_approximate_online_count = int_parser_factory('approximate_presence_count', 0)
put_approximate_online_count = int_putter_factory('approximate_presence_count')
validate_approximate_online_count = int_conditional_validator_factory(
    'approximate_online_count',
    0,
    lambda approximate_online_count : approximate_online_count >= 0,
    '>= 0',
)

# approximate_user_count

parse_approximate_user_count = int_parser_factory('approximate_member_count', 0)
put_approximate_user_count = int_putter_factory('approximate_member_count')
validate_approximate_user_count = int_conditional_validator_factory(
    'approximate_user_count',
    0,
    lambda approximate_user_count : approximate_user_count >= 0,
    '>= 0',
)

# channel

def parse_channel(data, guild_id = 0):
    """
    Parses a partial channel out of the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Invite data.
    guild_id : `int` = `0`, Optional
        The channel's guild's identifier.
    
    Returns
    -------
    channel : ``None | Channel``
    """
    channel_data = data.get('channel', None)
    if (channel_data is not None):
        return create_partial_channel_from_data(channel_data, guild_id)
    
    # If the data is partial, `channel_id` is received instead of `channel`.
    channel_id = data.get('channel_id', None)
    if (channel_id is not None):
        return create_partial_channel_from_id(int(channel_id), ChannelType.unknown, guild_id)
    
    return None


def put_channel(channel, data, defaults):
    """
    Puts the invite's channel into the given data.
    
    Parameters
    ----------
    channel : ``None | Channel``
        The channel to serialize.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if channel is None:
        channel_data = None
        channel_id = None
    else:
        channel_data = create_partial_channel_data(channel)
        channel_id = str(channel.id)
    
    data['channel'] = channel_data
    data['channel_id'] = channel_id
    
    return data


validate_channel = nullable_entity_validator_factory('channel', Channel)

# channel_id

put_channel_id = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)

# code

parse_code = force_string_parser_factory('code')
put_code = force_string_putter_factory('code')
validate_code = force_string_validator_factory('invite_code', 0, 1024)

# created_at

parse_created_at = nullable_date_time_parser_factory('created_at')
put_created_at = nullable_date_time_optional_putter_factory('created_at')
validate_created_at = nullable_date_time_validator_factory('created_at')

# flags

parse_flags = flag_parser_factory('flags', InviteFlag)
put_flags = flag_optional_putter_factory('flags', InviteFlag())
validate_flags = flag_validator_factory('flags', InviteFlag)

# guild

def parse_guild(data):
    """
    Parses a partial guild out of the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Invite data.
    
    Returns
    -------
    guild : ``None | Guild``
    """
    guild_data = data.get('guild', None)
    if (guild_data is not None):
        return create_partial_guild_from_data(guild_data)
    
    # If the data is partial, `guild_id` is received instead of `guild_id`.
    guild_id = data.get('guild_id', None)
    if (guild_id is not None):
        return create_partial_guild_from_id(int(guild_id))
    
    return None


def put_guild(guild, data, defaults):
    """
    Puts the invite's guild into the given data.
    
    Parameters
    ----------
    guild : ``None | Guild``
        The guild to serialize.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if guild is None:
        guild_data = None
        guild_id = None
    else:
        guild_data = create_partial_guild_data(guild)
        guild_id = str(guild.id)
    
    data['guild'] = guild_data
    data['guild_id'] = guild_id
    
    return data


validate_guild = nullable_entity_validator_factory('guild', Guild)


# guild_activity_overview

parse_guild_activity_overview = nullable_entity_parser_factory('profile', GuildActivityOverview)
put_guild_activity_overview = nullable_entity_optional_putter_factory(
    'profile', GuildActivityOverview, force_include_internals = True
)
validate_guild_activity_overview = nullable_entity_validator_factory('guild_activity_overview', GuildActivityOverview)


# guild_id

put_guild_id = entity_id_putter_factory('guild_id')

# inviter

parse_inviter = default_entity_parser_factory('inviter', User, default = ZEROUSER)
put_inviter = default_entity_putter_factory('inviter', ClientUserBase, ZEROUSER, force_include_internals = True)
validate_inviter = default_entity_validator_factory('inviter', ClientUserBase, default = ZEROUSER)

# inviter_id | extra for audit logs

validate_inviter_id = entity_id_validator_factory('inviter_id', ClientUserBase)

# max_age

parse_max_age = nullable_int_parser_factory('max_age')
put_max_age = nullable_field_optional_putter_factory('max_age')
validate_max_age = int_conditional_validator_factory(
    'max_age',
    None,
    lambda max_age : max_age >= 0,
    '>= 0',
)

# max_uses

parse_max_uses = nullable_int_parser_factory('max_uses')
put_max_uses = nullable_field_optional_putter_factory('max_uses')
validate_max_uses = int_conditional_validator_factory(
    'max_uses',
    None,
    lambda max_uses : max_uses >= 0,
    '>= 0',
)

# target_application

parse_target_application = nullable_functional_parser_factory('target_application', Application.from_data_invite)


def put_target_application(target_application, data, defaults):
    """
    Puts the invite's target application into the given data.
    
    Parameters
    ----------
    target_application : ``None | Application``
        The application to serialize.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (target_application is not None):
        if target_application is None:
            target_application_data = None
        else:
            target_application_data = target_application.to_data_invite(defaults = defaults, include_internals = True)
        
        data['target_application'] = target_application_data
    
    return data


validate_target_application = nullable_entity_validator_factory('target_application', Application)

# put_target_application_id

put_target_application_id = entity_id_optional_putter_factory('target_application_id')
validate_target_application_id = entity_id_validator_factory('target_application_id', Application)

# target_type

parse_target_type = preinstanced_parser_factory(
    'target_type', InviteTargetType, InviteTargetType.none
)


def put_target_type(target_type, data, defaults):
    """
    Puts the invite's target type into the given data.
    
    Parameters
    ----------
    target_type : ``InviteTargetType``
        The target type to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if target_type is not InviteTargetType.none:
        data['target_type'] = target_type.value
    
    return data


validate_target_type = preinstanced_validator_factory('target_type', InviteTargetType)

# target_user

parse_target_user = nullable_entity_parser_factory('target_user', User)
put_target_user = nullable_entity_optional_putter_factory(
    'target_user', ClientUserBase, force_include_internals = True
)
validate_target_user = nullable_entity_validator_factory('target_user', ClientUserBase)

# put_target_user_id

put_target_user_id = entity_id_optional_putter_factory('target_user_id')
validate_target_user_id = entity_id_validator_factory('target_user_id', ClientUserBase)

# temporary

parse_temporary = bool_parser_factory('temporary', False)
put_temporary = bool_optional_putter_factory('temporary', False)
validate_temporary = bool_validator_factory('temporary', False)

# type

parse_type = preinstanced_parser_factory('type', InviteType, InviteType.guild)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('invite_type', InviteType)

# validate_unique

put_unique = bool_optional_putter_factory('unique', False)
validate_unique = bool_validator_factory('unique', False)


# user_permissions

FIELDS_TO_PERMISSION_MASKS = (
    ('is_nickname_changeable', Permission.change_nickname.mask),
)

def parse_user_permissions(data):
    """
    Parses out the invite's user permissions from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    user_permissions : ``Permission``
    """
    permissions = 0
    
    for key, mask in FIELDS_TO_PERMISSION_MASKS:
        value = data.get(key, None)
        if (value is not None) and value:
            permissions |= mask
    
    return Permission(permissions)


def put_user_permissions(user_permissions, data, defaults):
    """
    Puts the invite's user permissions into the given data.
    
    Parameters
    ----------
    user_permissions : ``Permission``
        A subset of permissions that the user has upon joining.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    for key, mask in FIELDS_TO_PERMISSION_MASKS:
        value = True if user_permissions & mask else False
        if value or defaults:
            data[key] = value
    
    return data


validate_user_permissions = flag_validator_factory('user_permissions', Permission)


# uses

parse_uses = nullable_int_parser_factory('uses')
put_uses = nullable_field_optional_putter_factory('uses')
validate_uses = int_conditional_validator_factory(
    'uses',
    None,
    lambda uses : uses >= 0,
    '>= 0',
)



__all__ = ('SlashCommandFunction',)

from scarletio import RichAttributeErrorBaseType, copy_docs

from .....discord.application_command import ApplicationCommandOption, ApplicationCommandOptionType
from .....discord.client import Client
from .....discord.interaction import InteractionEvent

from ...converters import ParameterConverterInternal, ParameterConverterSlashCommand
from ...exceptions import handle_command_exception

from ...interfaces.autocomplete import AutocompleteInterface
from ...interfaces.command import CommandInterface
from ...interfaces.exception_handler import ExceptionHandlerInterface
from ...interfaces.self_reference import SelfReferenceInterface
from ...responding import process_command_coroutine

from ..command_base import CommandBase


class SlashCommandFunction(
    AutocompleteInterface,
    CommandInterface,
    ExceptionHandlerInterface,
    SelfReferenceInterface,
    RichAttributeErrorBaseType,
):
    """
    Represents a slash command's backend implementation.
    
    Attributes
    ----------
    _auto_completers : `None | list<SlashCommandParameterAutoCompleter>`
        Auto completer functions.
    
    _command_function : `async-callable˛
        The command's function to call.
    
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _parameter_converters : `tuple<ParameterConverterBase>`
        Parsers to parse command parameters.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>
        Reference to the parent application command or category.
    
    _self_reference : `None | WeakReferer<instance>`
        Back reference to the slasher application command function.
        
        Used by auto completers to access the parent entity.
    
    description : `str`
        The slash command's description.
    
    default : `bool`
        Whether the command is the default command in it's category.
    
    name : `str`
        The name of the slash command. It's length can be in range [1:32].
    
    response_modifier : `None | ResponseModifier`
        Modifies values returned and yielded to command coroutine processor.
    """
    __slots__ = (
        '__weakref__', '_auto_completers', '_command_function', '_exception_handlers', '_parameter_converters',
        '_parent_reference', '_self_reference', 'description', 'default', 'name', 'response_modifier'
    )
    
    def __new__(cls, function, parameter_converters, name, description, response_modifier, default):
        """
        Creates a new ``SlashCommandFunction`` with the given parameters.
        
        Parameters
        ----------
        function : `async-callable`
            The command's function to call.
        
        parameter_converters : `tuple<ParameterConverterBase>`
            Parsers to parse command parameters.
        
        name : `str`
            The name of the slash command.
        
        description : `str`
            The slash command's description.
        
        response_modifier : `None<ResponseModifier>`
            Modifies values returned and yielded to command coroutine processor.
        
        default : `bool`
            Whether the command is the default command in it's category.
        """
        self = object.__new__(cls)
        self._auto_completers = None
        self._command_function = function
        self._parameter_converters = parameter_converters
        self.response_modifier = response_modifier
        self.description = description
        self.name = name
        self.default = default
        self._exception_handlers = None
        self._parent_reference = None
        self._self_reference = None
        
        for parameter_converter in parameter_converters:
            parameter_converter.bind_parent(self)
        
        return self
    
    
    def __repr__(self):
        """Returns the application command option's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # description
        repr_parts.append(', description = ')
        repr_parts.append(repr(self.description))
        
        # default
        default = self.default
        if default:
            repr_parts.append(', default = ')
            repr_parts.append(repr(default))
        
        # response_modifier
        response_modifier = self.response_modifier
        if (response_modifier is not None):
            repr_parts.append(', response_modifier = ')
            repr_parts.append(repr(response_modifier))
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            repr_parts.append(', auto_completers = ')
            repr_parts.append(repr(auto_completers))
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            repr_parts.append(', exception_handlers = ')
            repr_parts.append(repr(exception_handlers))
        
        # _command_function
        repr_parts.append(', command_function = ')
        repr_parts.append(repr(self._command_function))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the slash command function's hash value."""
        hash_value = 0
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            hash_value ^= len(auto_completers)
            
            for auto_completer in auto_completers:
                hash_value ^= hash(auto_completer)
        
        # _command_function
        command_function = self._command_function
        try:
            command_function_hash_value = hash(command_function)
        except TypeError:
            command_function_hash_value = object.__hash__(command_function)
        hash_value ^= command_function_hash_value
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            hash_value ^= len(exception_handlers) << 4
            
            for exception_handler in exception_handlers:
                try:
                    exception_handler_hash_value = hash(exception_handler)
                except TypeError:
                    exception_handler_hash_value = object.__hash__(exception_handler)
                hash_value ^= exception_handler_hash_value
        
        # _parameter_converters
        # Internal field
        
        # _parent_reference
        # Internal field
        
        # _self_reference
        # Internal field
        
        # default
        hash_value ^= self.default << 8
        
        description = self.description
        hash_value ^= hash(description)
        
        # name
        name = self.name
        if name != description:
            hash_value ^= hash(name)
        
        # response_modifier
        hash_value ^= hash(self.response_modifier)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two slash command functions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # _auto_completers
        if self._auto_completers != other._auto_completers:
            return False
        
        # _command_function
        if self._command_function != other._command_function:
            return False
        
        # _exception_handlers
        if self._exception_handlers != other._exception_handlers:
            return False
        
        # _parameter_converters
        # Internal field
        
        # _parent_reference
        # Internal field
        
        # _self_reference
        # Internal field
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # response_modifier
        if self.response_modifier != other.response_modifier:
            return False
        
        return True
    
    
    def __format__(self, code):
        """Formats the slash command function in a format string."""
        if not code:
            return str(self)
        
        if code == 'm':
            return self.mention
        
        raise ValueError(
            f'Unknown format code {code!r} for {type(self).__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"m"!r}.'
        )
    
    
    async def invoke(self, client, interaction_event, options):
        """
        Calls the slash command function.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        
        interaction_event : ``InteractionEvent``
            The received interaction event.
        
        options : `None | list<InteractionEventChoice>``
            Options bound to the function.
        
        Raises
        ------
        SlashCommandParameterConversionError
            Exception occurred meanwhile parsing parameter.
        """
        parameters = []
        
        parameter_relation = {}
        if (options is not None):
            for option in options:
                parameter_relation[option.name] = option.value
        
        for parameter_converter in self._parameter_converters:
            if isinstance(parameter_converter, ParameterConverterInternal):
                value = None
            else:
                value = parameter_relation.get(parameter_converter.name, None)
            
            try:
                parameter = await parameter_converter(client, interaction_event, value)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                exception = err
            
            else:
                parameters.append(parameter)
                continue
            
            try:
                await handle_command_exception(
                    self,
                    client,
                    interaction_event,
                    exception,
                )
            finally:
                exception = None
            return
            
        
        command_coroutine = self._command_function(*parameters)
        
        try:
            await process_command_coroutine(
                client,
                interaction_event,
                self.response_modifier,
                command_coroutine,
            )
        except GeneratorExit:
            raise
        
        except BaseException as err:
            exception = err
        
        else:
            return
        
        try:
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
        finally:
            exception = None
        return
    
    
    async def invoke_auto_completion(self, client, interaction_event, auto_complete_option):
        """
        Calls the respective auto completion function of the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        
        interaction_event : ``InteractionEvent``
            The received interaction event.
        
        auto_complete_option : `InteractionMetadataApplicationCommandAutocomplete | InteractionOption`
            The option to autocomplete.
        """
        focused_option = auto_complete_option.focused_option
        if (focused_option is None):
            return
        
        parameter_name = focused_option.name
        
        for parameter_converter in self._parameter_converters:
            if isinstance(parameter_converter, ParameterConverterSlashCommand):
                if parameter_converter.name == parameter_name:
                    break
        else:
            return
        
        auto_completer = parameter_converter.auto_completer
        if (auto_completer is not None):
            await auto_completer.invoke(client, interaction_event)
    
    
    def copy(self):
        """
        Copies the slash command function.
        
        Returns
        -------
        self : ``SlashCommandFunction``
        """
        new = object.__new__(type(self))
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            auto_completers = auto_completers.copy()
        new._auto_completers = auto_completers
        
        # _command_function
        new._command_function = self._command_function
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        # _parameter_converters
        new._parameter_converters = self._parameter_converters
        
        # _parent_reference
        new._parent_reference = None
        
        # _self_reference
        new._self_reference = None
        
        # default
        new.default = self.default
        
        # description
        new.description = self.description
        
        # name
        new.name = self.name
        
        # response_modifier
        new.response_modifier = self.response_modifier
        
        return self
    
    
    @copy_docs(CommandInterface.get_command_function)
    def get_command_function(self):
        return self._command_function
    
    
    def _get_auto_completable_parameters(self):
        """
        Gets the auto completable parameter converters of the application command function.
        
        Returns
        -------
        parameter_converters : `list` of ``ParameterConverterSlashCommand``
        """
        auto_completable_parameters = set()
        
        for parameter_converter in self._parameter_converters:
            if (
                isinstance(parameter_converter, ParameterConverterSlashCommand) and
                parameter_converter.can_be_auto_completed()
            ):
                auto_completable_parameters.add(parameter_converter)
        
        return auto_completable_parameters
    
    
    @copy_docs(AutocompleteInterface._register_auto_completer)
    def _register_auto_completer(self, function, parameter_names):
        auto_completer = self._make_auto_completer(function, parameter_names)
        
        auto_completable_parameters = self._get_auto_completable_parameters()
        matched_auto_completable_parameters = auto_completer._difference_match_parameters(auto_completable_parameters)
        if not matched_auto_completable_parameters:
            raise RuntimeError(
                f'Application command function `{self.name}` has no parameter matching any '
                f'parameters of `{auto_completer!r}`.'
            )
        
        self._store_auto_completer(auto_completer)
        
        for parameter_converter in matched_auto_completable_parameters:
            parameter_converter.register_auto_completer(auto_completer)
        
        return auto_completer
    
    
    def _try_resolve_auto_completer(self, auto_completer):
        """
        Tries to register auto completer the instance.
        
        Parameters
        ----------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The auto completer.
        
        Returns
        -------
        resolved : `int`
            The amount of parameters resolved.
        """
        resolved = 0
        
        auto_completable_parameters = self._get_auto_completable_parameters()
        matched_auto_completable_parameters = auto_completer._difference_match_parameters(auto_completable_parameters)
        for parameter_converter in matched_auto_completable_parameters:
            resolved += parameter_converter.register_auto_completer(auto_completer)
        
        return resolved
    
    
    def as_option(self):
        """
        Returns the slash command function as an application command option.
        
        Returns
        -------
        option : ``ApplicationCommandOption``
        """
        parameter_converters = self._parameter_converters
        options = None
        for parameter_converter in parameter_converters:
            option = parameter_converter.as_option()
            if (option is not None):
                if options is None:
                    options = []
                
                options.append(option)
        
        return ApplicationCommandOption(
            self.name,
            self.description,
            ApplicationCommandOptionType.sub_command,
            options = options,
            default = self.default,
        )
    

    # ---- Mention ----
    
    @property
    @copy_docs(CommandBase.mention)
    def mention(self):
        parent_reference = self._parent_reference
        if parent_reference is None:
            parent = None
        else:
            parent = parent_reference()
        
        if parent is None:
            return ''
        
        return parent._mention_recursive(self.name)
    
    
    @copy_docs(CommandBase.mention_at)
    def mention_at(self, guild):
        parent_reference = self._parent_reference
        if parent_reference is None:
            parent = None
        else:
            parent = parent_reference()
        
        if parent is None:
            return ''
        
        return parent._mention_at_recursive(guild, self.name)



__all__ = ('BanEntry',)

from scarletio import RichAttributeErrorBaseType

from .fields import parse_reason, parse_user, put_reason, put_user, validate_reason, validate_user


class BanEntry(RichAttributeErrorBaseType):
    """
    A ban entry.
    
    Attributes
    ----------
    user : ``ClientUserBase``
        The banned user.
    reason : `None`, `str`
        The ban reason if applicable.
    """
    __slots__ = ('reason', 'user',)
    
    def __new__(cls, user, reason):
        """
        Creates a new ban entry instance.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The banned user.
        reason : `None`, `str`
            The ban reason if applicable.
        
        Raises
        -------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        reason = validate_reason(reason)
        user = validate_user(user)
        
        # Construct
        self = object.__new__(cls)
        self.reason = reason
        self.user = user
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a ban entry instance from the given `data`.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data representing a ban entry.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.reason = parse_reason(data)
        self.user = parse_user(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the ban entry into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_reason(self.reason, data, defaults)
        put_user(self.user, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the ban entry's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # user
        repr_parts.append(' user = ')
        repr_parts.append(repr(self.user))
        
        # reason
        reason = self.reason
        if (reason is not None):
            repr_parts.append(' reason = ')
            repr_parts.append(repr(reason))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __len__(self):
        """Helper for unpacking."""
        return 2
    
    
    def __iter__(self):
        """Unpacks the ban entry."""
        yield self.user
        yield self.reason

    
    def __hash__(self):
        """Returns the ban entry's hash value."""
        hash_value = 0
        
        # reason
        reason = self.reason
        if (reason is not None):
            hash_value ^= hash(reason)
        
        # user
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two ban entries are equal."""
        if type(self) is not type(other):
            return False
        
        # reason
        if self.reason != other.reason:
            return False
        
        # user
        if self.user != other.user:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the ban entry.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.reason = self.reason
        new.user = self.user
        return new
    
    
    def copy_with(self, *, reason = ..., user = ...):
        """
        Copies the ban entry.
        
        Parameters
        ----------
        reason : `None`, `str`, Optional (Keyword only)
            The ban reason if applicable.
        user : ``ClientUserBase``, Optional (Keyword only)
            The banned user.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        -------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # reason
        if reason is ...:
            reason = self.reason
        else:
            reason = validate_reason(reason)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # Construct
        new = object.__new__(type(self))
        new.reason = reason
        new.user = user
        return new



__all__ = ('ChannelMetadataGuildMedia',)

from .guild_forum_base import ChannelMetadataGuildForumBase


class ChannelMetadataGuildMedia(ChannelMetadataGuildForumBase):
    """
    Guild media channel metadata.
    
    Attributes
    ----------
    _cache_permission : ``None | dict<int, Permission>``
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    available_tags : `None`, `tuple` of ``ForumTag``
        The available tags to assign to the child-thread channels.
    default_forum_layout : ``ForumLayout``
        The default layout used to display threads of the forum.
    default_sort_order : ``SortOrder``
        The default thread ordering of the forum.
    default_thread_auto_archive_after : `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Defaults
        to `3600`. Can be one of: `3600`, `86400`, `259200`, `604800`.
    default_thread_reaction_emoji : ``None | Emoji``
        The emoji to show in the add reaction button on a thread of the forum channel.
    default_thread_slowmode : `int`
        The default slowmode applied to the channel's threads.
    flags : ``ChannelFlag``
        The channel's flags.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites : ``None | dict<int, PermissionOverwrite>``
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    topic : `None`, `str`
        The channel's topic.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()



__all__ = ('ReactionMapping', )

from scarletio import RichAttributeErrorBaseType

from ..emoji import Emoji, create_partial_emoji_data, create_partial_emoji_from_data
from ..reaction import Reaction, ReactionType
from ..reaction_mapping_line import ReactionMappingLine

from .fields import validate_lines


COUNT_KEYS_TO_TYPES = (
    ('normal', ReactionType.standard),
    ('burst', ReactionType.burst),
)


class ReactionMapping(RichAttributeErrorBaseType):
    """
    Contains the reactions on a ``Message`` with (``Reaction``, ``ReactionMappingLine``) items.
    
    Attributes
    ----------
    lines : `None | dict<Reaction, ReactionMappingLine>`
        Reaction to users relation.
    """
    __slots__ = ('lines',)
    
    def __new__(cls, *, lines = ...):
        """
        Creates a new reaction mapping instance.
        
        Parameters
        ----------
        lines : `None | dict<str | Emoji | Reaction, ReactionMappingLine> \
                | list<(str | Emoji | Reaction, ReactionMappingLine)>`, Optional (Keyword only)
            Reaction to users relation.
        
        Raises
        ------
        TypeError
            - If `initialize_with`'s is unacceptable.
        """
        if lines is ...:
            lines = None
        else:
            lines = validate_lines(lines)
        
        # Construct
        self = object.__new__(cls)
        self.lines = lines
        return self
    
    
    def __len__(self):
        """Returns the length of the reaction mapping."""
        lines = self.lines
        if lines is None:
            return 0
        
        return len(lines)
    
    
    def __bool__(self):
        """Returns whether self has any any reactions"""
        
        # lines
        lines = self.lines
        if (lines is not None) and lines:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the reaction mapping's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # lines
        lines = self.lines
        if (lines is not None) and lines:
            repr_parts.append('lines = ')
            repr_parts.append(repr(lines))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether self equals to other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # lines
        self_lines = self.lines
        other_lines = other.lines
        if (self_lines is None) or (not self_lines):
            if (other_lines is not None) and other_lines:
                return False
        
        else:
            if (other_lines is None) or (not other_lines):
                return False
            
            if self_lines != other_lines:
                return False
        
        return True
    
    
    def __hash__(self):
        """
        Returns the hash value of reaction mapping line.
        > Note that it is the current hash of the object and can change by changing the object.
        """
        hash_value = 0
        
        lines = self.lines
        if (lines is not None):
            for reaction, line in lines.items():
             hash_value ^= hash(reaction) & hash(line)
        
        return hash_value
    
    
    @classmethod
    def from_data(cls, data):
        """
        Fills the reaction mapping with the given data.
        
        Parameters
        ----------
        data : `None`, `list<dict<str, object>>`
            Reactions data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        lines = None
        
        for line_data in data:
            counts = line_data.get('count_details', None)
            if counts is None:
                continue
            
            emoji = create_partial_emoji_from_data(line_data['emoji'])
            
            for key, reaction_type in COUNT_KEYS_TO_TYPES:
                count = counts.get(key, 0)
                if not count:
                    continue
                
                if lines is None:
                    lines = {}
                
                line = ReactionMappingLine._create_empty()
                line.count = count
                lines[Reaction.from_fields(emoji, reaction_type)] = line
        
        # Construct
        self = object.__new__(cls)
        self.lines = lines
        return self
    
    
    def to_data(self):
        """
        Tries to convert the reactions back to a json serializable list.
        
        Returns
        -------
        data : `list<dict<str, object>>`
        """
        data = []
        
        lines = self.lines
        if (lines is not None):
            reduced_to_emojis = {}
            for reaction, line in lines.items():
                emoji = reaction.emoji
                try:
                    by_type = reduced_to_emojis[emoji]
                except KeyError:
                    by_type = {}
                    reduced_to_emojis[emoji] = by_type
                
                by_type[reaction.type] = line
                
                
            for emoji, by_type in reduced_to_emojis.items():
                counts = {}
                
                for key, reaction_type in COUNT_KEYS_TO_TYPES:
                    line = by_type.get(reaction_type, None)
                    counts[key] = 0 if (line is None) else line.count
                
                data.append({
                    'count_details': counts,
                    'emoji': create_partial_emoji_data(emoji),
                })
            
        return data
    
    
    @property
    def emoji_count(self):
        """
        The amount of different emojis, which were added on the reaction mapping's respective ``Message``.
        
        Returns
        -------
        emoji_count : `int`
        """
        lines = self.lines
        if lines is None:
            return 0
        
        return len({reaction.emoji for reaction in lines.keys()})
    
    
    @property
    def reaction_count(self):
        """
        The amount of different reactions, which were added on the reaction mapping's respective ``Message``.
        
        Returns
        -------
        reaction_count : `int`
        """
        lines = self.lines
        if lines is None:
            return 0
        
        return len(lines)
    
    
    @property
    def total_count(self):
        """
        The total amount reactions given on the reaction mapping's respective message.
        
        Returns
        -------
        total_count : `int`
        """
        lines = self.lines
        if lines is None:
            return 0
        
        return sum(line.count for line in lines.values())
    
    
    def copy(self):
        """
        Copies the reaction mapping returning a new one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # lines
        lines = self.lines
        if (lines is not None):
            lines = {reaction.copy(): line.copy() for reaction, line in lines.items()}
        new.lines = lines
        
        return new
    
    
    def _add_reaction(self, reaction, user):
        """
        Adds a user to the reactors.
        
        Parameters
        ----------
        reaction : ``Reaction``
            The reaction.
        user : ``ClientUserBase``
            The reactor user.
        
        Returns
        -------
        success : `bool`
        """
        line = self._get_or_create_line(reaction)
        return line._add_reaction(user)
    
    
    def _remove_reaction(self, reaction, user):
        """
        Removes a user from the reactors.
        
        Parameters
        ----------
        reaction : ``Reaction``
            The reaction.
        user : ``ClientUserBase``
            The removed reactor user.
        
        Returns
        -------
        success : `bool`
        """
        lines = self.lines
        if lines is None:
            return False
        
        try:
            line = lines[reaction]
        except KeyError:
            return False
        
        success = line._remove_reaction(user)
        if success:
            if not line.count:
                del lines[reaction]
        
        return success
    
    
    def _remove_reaction_emoji(self, emoji):
        """
        Removes all the users who reacted with the given ``Emoji`` and then returns the stored line.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji to remove.
        
        Returns
        -------
        removed_lines : `None | dict<Reaction, ReactionMappingLine>`
        """
        removed_lines = None
        
        lines = self.lines
        if (lines is not None):
            for reaction, line in lines.items():
                if reaction.emoji is not emoji:
                    continue
            
                if removed_lines is None:
                    removed_lines = {}
            
                removed_lines[reaction] = line
                continue
        
            if (removed_lines is not None):
                for reaction in removed_lines.keys():
                    del lines[reaction]
            
        return removed_lines
    
    
    def _fill_some_reactions(self, reaction, users):
        """
        Called when some reactors of a reaction are updated.
        
        Parameters
        ----------
        reaction : ``Reaction``
            The reaction, which users' are updated.
        users : ``list<ClientUserBase>``
            The added reactors.
        """
        line = self._get_or_create_line(reaction)
        line._fill_some_reactions(users)
    
    
    def _fill_all_reactions(self, reaction, users):
        """
        Called when all the reactors of a reaction are updated of the reaction mapping.
        
        Parameters
        ----------
        reaction : ``Reaction``
            The reaction, which users' are updated.
        users : ``list<ClientUserBase>``
            The added reactors.
        """
        line = self._get_or_create_line(reaction)
        line._fill_all_reactions(users)
    
    
    def __getitem__(self, reaction):
        """
        Gets the result for the given answer.
        
        Parameters
        ----------
        reaction : `Emoji | Reaction`
            
        Returns
        -------
        result : `None | ReactionMappingLine`
       
        """
        lines = self.lines
        if (lines is not None):
            return lines.get(reaction, None)
    
    
    def _get_or_create_line(self, reaction):
        """
        Gets the line for the given `reaction`. If not found creates a new one.
        
        Parameters
        ----------
        reaction : ``Reaction``
            The reaction pointing at the line.
        
        Returns
        -------
        line : ``ReactionMappingLine``
        """
        lines = self.lines
        if lines is None:
            line = ReactionMappingLine._create_empty()
            self.lines = {reaction: line}
        else:
            try:
                line = lines[reaction]
            except KeyError:
                line = ReactionMappingLine._create_empty()
                lines[reaction] = line
        
        return line
    
    
    def clear(self):
        """
        Clears the non-client users of the reaction mapping.
        """
        lines = self.lines
        if (lines is not None):
            lines.clear()
    
    
    def __contains__(self, reaction):
        """Returns whether the reaction mapping has the given reaction."""
        lines = self.lines
        if lines is None:
            return False
        
        return reaction in lines
    
    
    def iter_reactions(self):
        """
        Iterates over reactions.
        
        This method is an iterable generator.
        
        Yields
        ------
        reaction : ``Reaction``
        """
        lines = self.lines
        if (lines is not None):
            yield from lines.keys()
    
    
    def iter_lines(self):
        """
        Iterates over lines.
        
        This method is an iterable generator.
        
        Yields
        ------
        item : ``ReactionMappingLine``
        """
        lines = self.lines
        if (lines is not None):
            yield from lines.values()
    
    
    def iter_items(self):
        """
        Iterates over `reaction - lines` pairs.
        
        This method is an iterable generator.
        
        Yields
        ------
        item : `(Reaction, ReactionMappingLine)`
        """
        lines = self.lines
        if (lines is not None):
            yield from lines.items()



__all__ = ('AudioSource', 'DownloadError', 'LocalAudio', 'RawAudio', 'YTAudio')

import os, subprocess
from os.path import isfile as is_file
from pathlib import Path
from shlex import split

from scarletio import CancelledError, Task, alchemy_incendiary, copy_docs

from ..core import KOKORO

from .audio_settings import AUDIO_SETTINGS_DEFAULT


DEFAULT_EXECUTABLE = 'ffmpeg'

if os.name == 'nt':
    SUBPROCESS_STARTUP_INFO = subprocess.STARTUPINFO()
    SUBPROCESS_STARTUP_INFO.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    SUBPROCESS_STARTUP_INFO.wShowWindow = subprocess.SW_HIDE
else:
    SUBPROCESS_STARTUP_INFO = None

STREAM_OPTIONS = (
    '-reconnect', '1',
    # '-reconnect_streamed', '1',
    # '-reconnect_delay_max', '3',
)

class AudioSource:
    """
    Base class for audio sources.
    
    Class Attributes
    ----------------
    AUDIO_SETTINGS : ``AudioSettings`` = `AUDIO_SETTINGS_DEFAULT`
        Settings containing how the audio should be played.
    NEEDS_ENCODE : `bool` = `True`
        Whether the source is not opus encoded.
    REPEATABLE : `bool` = `False`
        Whether the source can be repeated after it is exhausted once.
    """
    __slots__ = ()
    
    AUDIO_SETTINGS = AUDIO_SETTINGS_DEFAULT
    NEEDS_ENCODE = True
    REPEATABLE = False
    
    async def read(self):
        """
        Reads a chunk of data. by default should be 20ms.
        
        Indicates end of stream by returning `None`.
        
        Subclasses should implement it.
        
        This method is a coroutine.
        
        Returns
        -------
        audio_data : `bytes`, `None`
        """
        return None
    
    
    async def cleanup(self):
        """
        Cleans up the allocated resources by the audio source.
        
        Subclasses should overwrite it.
        
        This method is a coroutine.
        """
        pass
    
    
    def __del__(self):
        """Cleans up the audio source if ``.cleanup`` was not called for any reason."""
        Task(KOKORO, self.cleanup())
    
    
    @property
    def title(self):
        """
        Placeholder method for title attribute.
        
        Always returns an empty string.
        
        Returns
        -------
        title : `str`
        """
        return ''
    
    
    @property
    def path(self):
        """
        Placeholder method for path attribute.
        
        Always returns `None`.
        
        Returns
        -------
        path : `None`
        """
        return None
    
    
    async def postprocess(self):
        """
        Called before the audio of the source would be played.
        
        This method is a coroutine.
        """
        pass


class RawAudio(AudioSource):
    """
    Base type for raw audio data (`bytes`).
    
    Attributes
    ----------
    data : `bytes`
        Raw audio data.
    length : `int`
        The length of `data`.
    position : `int`
        The next read position.
    
    Class Attributes
    ----------------
    AUDIO_SETTINGS : ``AudioSettings`` = `AUDIO_SETTINGS_DEFAULT`
        Settings containing how the audio should be played.
    NEEDS_ENCODE : `bool` = `True`
        Whether the source is not opus encoded.
    REPEATABLE : `bool` = `False`
        Whether the source can be repeated after it is exhausted once.
    
    Examples
    --------
    In practice you probably want to subclass this type, because you raw data might not match the expected format.
    
    Lets say we have `f32` numpy array input with 24000 sampling rate and `1` channel.
    ```py
    from numpy import int16
    from hata import RawAudio
    
    class GeneratedAudio(RawAudio):
        __slots__ = ()
        
        AUDIO_SETTINGS = RawAudio.AUDIO_SETTINGS.copy_with(
            channels = 1,
            sampling_rate = 24000,
        )
        
        def __new__(cls, f32_array):
            # First we convert the `f32` array to `i16`. But there is a catch!
            # The f32 array will probably have values from `-1` to `+1`, so first we need normalise them.
            data = ((f32_array * (1 << 15)).astype(int16)).tobytes()
            return RawAudio.__new__(cls, data)
    ```
    """
    __slots__ = ('data', 'length', 'position')
    
    AUDIO_SETTINGS = AUDIO_SETTINGS_DEFAULT
    NEEDS_ENCODE = True
    REPEATABLE = True
    
    def __new__(cls, data):
        """
        Creates a new raw audio source.
        
        Parameters
        ----------
        data : `bytes`
            Raw audio data.
        """
        self = object.__new__(cls)
        self.data = data
        self.length = len(data)
        self.position = 0
        return self
    
    
    @copy_docs(AudioSource.__del__)
    def __del__(self):
        pass
    
    
    @copy_docs(AudioSource.read)
    async def read(self):
        chunk_start = self.position
      
        if chunk_start < self.length:
            chunk_end = chunk_start + self.AUDIO_SETTINGS.frame_size
            self.position = chunk_end
            return self.data[chunk_start : chunk_end]
    
    @copy_docs(AudioSource.postprocess)
    async def postprocess(self):
        # Reset position in case we want to repeat it.
        self.position = 0


class LocalAudio(AudioSource):
    """
    Represents a ffmpeg pcm audio.
    
    You must have the ffmpeg or avconv executable in your path environment variable in order for this to work.
    
    Attributes
    ----------
    _process_args : `tuple` ((`list` of `str`),  (`None`, `file-like`))
        Parameters and the stdin used to open the postprocess when postprocess happens.
    _stdout : `_io.BufferedReader`
        Stdout of `.process`.
    path : `None`, `str`
        The audio source's path if applicable. Defaults to `None`.
    process : `subprocess.Popen`
        The ffmpeg or the avconv subprocess.
    title : `str`
        The audio source's title if applicable. Defaults to empty string.
    
    Class Attributes
    ----------------
    AUDIO_SETTINGS : ``AudioSettings`` = `AUDIO_SETTINGS_DEFAULT`
        Settings containing how the audio should be played.
    NEEDS_ENCODE : `bool` = `True`
        Whether the source is not opus encoded.
    REPEATABLE : `bool` = `True`
        Whether the source can be repeated after it is exhausted once.
    """
    REPEATABLE = True
    
    @classmethod
    def _create_process_preprocess(cls, source, executable, pipe, before_options, options):
        """
        Creates a subprocess' args to open.
        
        Parameters
        ----------
        source : `str`, `Path`, `file-like`
            The source audio file's path or `file-like` if `pipe` is `True`.
        executable : `str`
            The executable's name to use. Defaults to `'ffmpeg'`.
        pipe : `bool`
            Whether the source is passed to stdin.
        before_options : `str` or (`iterable` of `str`)
            Extra parameters passed before the `-i` flag.
        options : `str` or (`iterable` of `str`)
            Extra parameters passed after the `-i` flag.
        
        Returns
        -------
        executable : `str`
            The executable's name.
        args : `list` of `str`
            Subprocess parameters.
        stdin : `None or `file-like`
            Input for the postprocess.
        
        Raises
        ------
        TypeError
            - If `pipe` was given as `True` meanwhile `source` was not given as a `file-like` supporting `.fileno()`
                method.
            - If `pipe` was given as `False`, meanwhile `source` was not given as `str`, `Path`.
        ValueError
            - Executable as not found.
            - Popen failed.
            - File does not exist.
        """
        if pipe:
            try:
                fileno_function = type(source).fileno
            except AttributeError as err:
                raise TypeError(
                    f'The given `source` not supports `.fileno()` method, got {source!r}.'
                ) from err
            try:
                fileno_function(source)
            except TypeError as err:
                raise TypeError(
                    f'The given `source` not supports `.fileno()` method, got {source!r}.'
                ) from err
        else:
            source_type = type(source)
            if source_type is str:
                pass
            elif issubclass(source_type, Path):
                source = str(source)
            elif issubclass(source_type, str):
                source = str(source)
            else:
                raise TypeError(
                    f'`source` can be `str`, `Path`, got {source_type.__name__}; {source!r}.'
                )
            
            if not is_file(source):
                raise ValueError(
                    f'`source` is not a file. Got: {source!r}',
                )
        
        
        args = []
        
        if (before_options is not None):
            if isinstance(before_options, str):
                before_options = split(before_options)
            
            args.extend(before_options)
        
        audio_settings = cls.AUDIO_SETTINGS
        
        args.append('-i')
        args.append('-' if pipe else source)
        args.append('-f')
        args.append('s16le')
        args.append('-ar')
        args.append(str(audio_settings.sampling_rate))
        args.append('-ac')
        args.append(str(audio_settings.channels))
        args.append('-loglevel')
        args.append('panic')
        
        if (options is not None):
            if isinstance(options, str):
                options = split(options)
            
            args.extend(options)
        
        args.append('pipe:1')
        
        return executable, args, (source if pipe else None)
    
    
    async def postprocess(self):
        """
        Creates the process of the audio player.
        
        This method is a coroutine.
        
        Raises
        ------
        ValueError
            - Executable as not found.
            - Popen failed.
        """
        process = self.process
        if process is None:
            executable, args, stdin = self._process_args
            try:
                process = await KOKORO.subprocess_exec(
                    executable, *args, stdin = stdin, stdout = subprocess.PIPE, startup_info = SUBPROCESS_STARTUP_INFO
                )
            except FileNotFoundError:
                raise ValueError(
                    f'{executable!r} was not found.'
                ) from None
            
            except subprocess.SubprocessError as err:
                raise ValueError(
                    f'Opening subprocess failed: {err.__class__.__name__}: {err}'
                ) from err
            
            self.process = process
            self._stdout = process.stdout
    
    __slots__ = ('_process_args', '_stdout', 'path', 'process', 'title', )
    
    # use __new__, so __del__ wont run
    async def __new__(
        cls,
        source,
        executable = DEFAULT_EXECUTABLE,
        pipe = False,
        before_options = None,
        options = None,
        title = None,
    ):
        """
        Creates a new ``LocalAudio``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        source : `str`, `file-like`
            The source audio file's path or `file-like` if `pipe` is `True`.
        
        executable : `str` = `DEFAULT_EXECUTABLE`, Optional
            The executable's name to use. Defaults to `'ffmpeg'`.
        
        pipe : `bool` = `False`, Optional
            Whether the source is passed to stdin.
        
        before_options : `str`, (`iterable` of `str`) = `None`, Optional
            Extra parameters passed before the `-i` flag.
        
        options : `str`, (`iterable` of `str`) = `None`, Optional
            Extra parameters passed after the `-i` flag.
        
        Returns
        -------
        self : ``LocalAudio``
        
        Raises
        ------
        TypeError
            - If `pipe` was given as `True` meanwhile `source` was not given as a `file-like` supporting `.fileno()`
                method.
            - If `pipe` was given as `False`, meanwhile `source` was not given as `str`, `Path`.
        """
        args = cls._create_process_preprocess(source, executable, pipe, before_options, options)
        
        self = object.__new__(cls)
        self._process_args = args
        self.process = None
        self._stdout = None
        
        if pipe:
            path = None
            if title is None:
                title = getattr(source, 'name', None)
                if title is None:
                    title = ''
                else:
                    title = os.path.splitext(title)[0].replace('_', ' ')
        else:
            path = source
            if title is None:
                title = os.path.splitext(os.path.basename(path))[0].replace('_', ' ')
        
        self.path = path
        self.title = title
        
        return self
    
    async def read(self):
        """
        Reads 20ms audio data.
        
        Indicates end of stream by returning zero `None`.
        
        This method is a coroutine.
        
        Returns
        -------
        audio_data : `bytes`, `None`
        """
        stdout = self._stdout
        if stdout is None:
            result = None
        else:
            frame_size = self.AUDIO_SETTINGS.frame_size
            try:
                result = await stdout.read(frame_size)
            except (CancelledError, ConnectionError):
                result = None
            else:
                if len(result) != frame_size:
                    result = None
        
        return result
    
    async def cleanup(self):
        """
        Closes ``.process``.
        
        This method is a coroutine.
        """
        process = self.process
        if process is None:
            return
        
        await process.kill()
        if process.poll() is None:
            await process.communicate()
        
        self._stdout = None
        self.process = None


try:
    import youtube_dl
except ImportError:
    youtube_dl = None
    DownloadError = None
    YTAudio = None
else:
    from youtube_dl.utils import DownloadError
    youtube_dl.utils.bug_reports_message = lambda: ''
    
    YTdl = youtube_dl.YoutubeDL({
        'format'            : 'bestaudio/best',
        'outtmpl'           : '%(extractor)s-%(id)s-%(title)s.%(ext)s',
        'restrictfilenames' : True,
        'noplaylist'        : True,
        'nocheckcertificate': True,
        'ignoreerrors'      : False,
        'logtostderr'       : False,
        'quiet'             : True,
        'no_warnings'       : True,
        'default_search'    : 'auto',
        'source_address'    : '0.0.0.0',
    })
    
    class YTAudio(LocalAudio):
        """
        Represents an audio sourced downloaded from youtube.
        
        You must have the ffmpeg or avconv executable in your path environment variable in order for this to work.
        
        Attributes
        ----------
        _process_args : `tuple` ((`list` of `str`),  (`None`, `file-like`))
            Parameters and the stdin used to open the postprocess when postprocess happens.
        _stdout : `_io.BufferedReader`
            Stdout of `.process`.
        path : `None`, `str`
            The audio source's path if applicable. Defaults to `None`.
        process : `subprocess.Popen`
            The ffmpeg or the avconv subprocess.
        title : `str`
            The audio source's title if applicable. Defaults to empty string.
        url : `str`
            The source url of the downloaded audio.
            
        Class Attributes
        ----------------
        AUDIO_SETTINGS : ``AudioSettings`` = `AUDIO_SETTINGS_DEFAULT`
            Settings containing how the audio should be played.
        NEEDS_ENCODE : `bool` = `True`
            Whether the source is not opus encoded.
        REPEATABLE : `bool` = `True`
            Whether the source can be repeated after it is exhausted once.
        """
        
        __slots__ = ('url', )
        
        @staticmethod
        def _preprocess(cls, url, stream):
            """
            Downloads the audio source by the given url or title.
            
            This function runs inside of an executor thread.
            
            Parameters
            ----------
            url : `str`
            
            Returns
            -------
            path : `str`
                The title of the downloaded audio.
            data : `dict<str, object>`
                All extracted data by YTDL.
            args : `list` of `str`
                Subprocess parameters.
            
            Raises
            ------
            DownloadError
                Downloading the audio source failed.
            """
            data = YTdl.extract_info(url, download = (not stream))
            
            if 'entries' in data: #playlist
                data = data['entries'][0]
            
            if stream:
                path = data['url']
                before_options = STREAM_OPTIONS
            else:
                path = YTdl.prepare_filename(data)
                before_options = None
            
            args = cls._create_process_preprocess(path, DEFAULT_EXECUTABLE, False, before_options, ('-vn',))
            
            return path, data, args
        
        
        async def __new__(cls, url, stream = True):
            """
            Creates a new ``YTAudio``.
            
            This method is a coroutine.
            
            Parameters
            ----------
            url : `str`
                The url or the title of the video.
            stream : `bool` = `True`, Optional
                Whether the audio should be streamed.
            
            Returns
            -------
            self : ``YTAudio``
            
            Raises
            ------
            DownloadError
                Downloading the audio source failed.
            PermissionError
                The given file started to be played at the same time by an other player as well.
            TypeError
                - If `pipe` was given as `True` meanwhile `source` was not given as a `file-like` supporting `.fileno()`
                    method.
                - If `pipe` was given as `False`, meanwhile `source` was not given as `str`, `Path`.
            """
            path, data, args = await KOKORO.run_in_executor(alchemy_incendiary(cls._preprocess, (cls, url, stream)))
            
            # Create self only at the end, so the `__del__` wont pick it up
            self = object.__new__(cls)
            self._process_args = args
            self.process = None
            self._stdout = None
            self.path = path
            self.title = data.get('title', None)
            self.url = data.get('url', None)
            
            return self



__all__ = ('InteractionMetadataMessageComponent',)

from warnings import warn

from scarletio import copy_docs

from ...component import Component, ComponentType, InteractionComponent

from .base import InteractionMetadataBase
from .fields import parse_component, validate_component, put_component


class InteractionMetadataMessageComponent(InteractionMetadataBase):
    """
    Interaction metadata used when the interaction was triggered by an application command's auto completion.
    
    Attributes
    ----------
    component : ``None | InteractionComponent``
        The interacted interaction component.
    """
    __slots__ = ('component',)
    
    def __new__(
        cls,
        *,
        component = ...,
        component_type = ...,
        custom_id = ...,
        values = ...,
    ):
        """
        Creates a new interaction metadata from the given parameters.
        
        Parameters
        ----------
        component : ``None | InteractionComponent``, Optional (Keyword only)
            The interacted component.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # Deprecations
        if (component_type is not ...) or (custom_id is not ...) or (values is not ...):
            if (component_type is not ...):
                warn(
                    (
                        f'`{cls.__name__}.__new__`\'s `component_type` parameter is deprecated and is scheduled for '
                        f'removal at 2026 February. Please use the `component` parameter instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            if (custom_id is not ...):
                warn(
                    (
                        f'`{cls.__name__}.__new__`\'s `custom_id` parameter is deprecated and is scheduled for '
                        f'removal at 2026 February. Please use the `component` parameter instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            if (values is not ...):
                warn(
                    (
                        f'`{cls.__name__}.__new__`\'s `values` parameter is deprecated and is scheduled for '
                        f'removal at 2026 February. Please use the `component` parameter instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            component = InteractionComponent(
                **{'component_type': ComponentType.string_select if component_type is ... else component_type},
                **({} if custom_id is ... else {'custom_id': custom_id}),
                **({} if values is ... else {'values': values}),
            )
        
        
        # component
        if component is ...:
            component = None
        else:
            component = validate_component(component)
        
        # Construct
        self = object.__new__(cls)
        self.component = component
        return self
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            component = keyword_parameters.pop('component', ...),
        )
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.component = None
        return self
    
    
    @copy_docs(InteractionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        component = self.component
        if (component is not None):
            component = component.copy()
        new.component = component
        
        return new
    
    
    def copy_with(
        self,
        *,
        component = ...,
        component_type = ...,
        custom_id = ...,
        values = ...,
    ):
        """
        Copies the interaction metadata with the given fields.
        
        Parameters
        ----------
        component : ``None | InteractionComponent``, Optional (Keyword only)
            The interacted component.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # Deprecations
        if (component_type is not ...) or (custom_id is not ...) or (values is not ...):
            if (component_type is not ...):
                warn(
                    (
                        f'`{type(self).__name__}.copy_with`\'s `component_type` parameter is deprecated and is '
                        f'scheduled for removal at 2026 February. Please use the `component` parameter instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            if (custom_id is not ...):
                warn(
                    (
                        f'`{type(self).__name__}.copy_with`\'s `custom_id` parameter is deprecated and is scheduled '
                        f'for removal at 2026 February. Please use the `component` parameter instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            if (values is not ...):
                warn(
                    (
                        f'`{type(self).__name__}.copy_with`\'s `values` parameter is deprecated and is scheduled for '
                        f'removal at 2026 February. Please use the `component` parameter instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            component = InteractionComponent(
                **{'component_type': ComponentType.string_select if component_type is ... else component_type},
                **({} if custom_id is ... else {'custom_id': custom_id}),
                **({} if values is ... else {'values': values}),
            )
        
        # component
        if component is ...:
            component = self.component
            if (component is not None):
                component = component.copy()
        else:
            component = validate_component(component)
        
        # Construct
        new = object.__new__(type(self))
        new.component = component
        return new
    
    
    @copy_docs(InteractionMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            component = keyword_parameters.pop('component', ...),
        )
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_data)
    def from_data(cls, data, guild_id = 0):
        self = object.__new__(cls)
        self.component = parse_component(data)
        return self
    
    
    @copy_docs(InteractionMetadataBase.to_data)
    def to_data(self, *, defaults = False, guild_id = 0):
        data = {}
        put_component(self.component, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataBase._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        # component
        repr_parts.append(' component = ')
        repr_parts.append(repr(self.component))
        return True
    
    
    @copy_docs(InteractionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # component
        hash_value ^= hash(self.component)
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataBase.__eq__)
    def __eq__(self, other):
        other_type = type(other)
        if other_type is type(self):
            return self._is_equal_same_type(other)
        
        # Compare with components.
        if issubclass(other_type, Component):
            component = self.component
            if component is None:
                return False
            
            # Check `type` before `custom_id`
            
            # type
            if component.type is not other.type:
                return False
            
            # custom_id
            if component.custom_id != other.custom_id:
                return False
            
            return True
        
        return NotImplemented
    
    
    @copy_docs(InteractionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # component
        if self.component != other.component:
            return False
        
        return True
    
    
    @copy_docs(InteractionMetadataBase.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        component = self.component
        if (component is not None):
            yield from component.iter_custom_ids_and_values()
    
    
    @property
    @copy_docs(InteractionMetadataBase.custom_id)
    def custom_id(self):
        component = self.component
        if component is not None:
            return component.custom_id



__all__ = ('VoiceAttachment',)

from scarletio import RichAttributeErrorBaseType

from .constants import WAVEFORM_OGG_DEFAULT
from .fields import (
    put_description, put_duration, put_waveform, validate_description, validate_duration, validate_name,
    validate_waveform
)


class VoiceAttachment(RichAttributeErrorBaseType):
    """
    Voice attachment to be attachable to a message.
    
    Attributes
    ----------
    description : `None | str`
        Description for the attachment.
    
    duration : `float`
        The attachment's duration in seconds.
    
    io : `object`
        Data or stream to be sent.
    
    name : `str`
        The name of the attachment.
    
    waveform : `bytes`
        Represents a sampled waveform of the attached voice data.
    """
    __slots__ = ('description', 'duration', 'io', 'name', 'waveform')
    
    def __new__(cls, name, io, duration, *, description = ..., waveform = ...):
        """
        Creates a new voice attachment.
        
        Parameters
        ----------
        name : `str`
            The name of the attachment.
        
        io : `object`
            Data or stream to be sent.
        
        duration : `float`
            The attachment's duration in seconds.
        
        description : `None | str`, Optional (Keyword only)
            Description for the attachment.
        
        waveform : `None | bytes`, Optional (Keyword only)
            Represents a sampled waveform of the attached voice data.
        
        Raises
        ------
        ValueError
            - If a parameter's value is incorrect.
            - If `waveform` is not given and could not interpret the default waveform.
        TypeError
            - If a parameter's type is incorrect.
        """
        # name
        name = validate_name(name)
        
        # duration
        duration = validate_duration(duration)
        
        # description
        if description is ...:
            description = None
        
        else:
            description = validate_description(description)
        
        # waveform
        if waveform is ...:
            waveform = None
        else:
            waveform = validate_waveform(waveform)
        
        # Post validate waveform.
        if waveform is None:
            if name.endswith('.ogg'):
                waveform = WAVEFORM_OGG_DEFAULT
            
            else:
                raise ValueError(
                    f'Could not interpret default `waveform` from `name` ({name!r}), please pass it manually.'
                )
        
        # Construct
        self = object.__new__(cls)
        self.description = description
        self.duration = duration
        self.io = io
        self.name = name
        self.waveform = waveform
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # duration
        repr_parts.append(', duration = ')
        repr_parts.append(repr(self.duration))
        
        # type(io)
        repr_parts.append(', type(io) = ')
        repr_parts.append(type(self.io).__name__)
        
        # description
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(repr(description))
        
        # waveform
        repr_parts.append(', waveform = ')
        repr_parts.append(repr(self.waveform))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # duration
        hash_value ^= hash(self.duration)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # io
        io = self.io
        try:
            io_hash = hash(io)
        except TypeError:
            io_hash = object.__hash__(io)
        
        hash_value ^= io_hash
        
        # name
        hash_value ^= hash(self.name)
        
        # waveform
        hash_value ^= hash(self.waveform)
        
        return hash_value
        
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # description
        if self.description != other.description:
            return False
        
        # duration
        if self.duration != other.duration:
            return False
        
        # io
        if self.io != other.io:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # waveform:
        if self.waveform != other.waveform:
            return False
        
        return True
    
    
    def copy(self):
        """
        Returns a copy of the voice attachment.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.description = self.description
        new.duration = self.duration
        new.io = self.io
        new.name = self.name
        new.waveform = self.waveform
        return new
    
    
    def copy_with(self, *, description = ..., duration = ..., io = ..., name = ..., waveform = ...):
        """
        Copies the voice attachment with the given fields.
        
        Parameters
        ----------
        description : `None | str`, Optional (Keyword only)
            Description for the attachment.
        
        duration : `float`, Optional (Keyword only)
            The attachment's duration in seconds.
        
        io : `object`, Optional (Keyword only)
            Data or stream to be sent.
        
        name : `str`, Optional (Keyword only)
            The name of the attachment.
        
        waveform : `None | bytes`, Optional (Keyword only)
            Represents a sampled waveform of the attached voice data.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            - If a parameter's value is incorrect.
            - If `waveform` is not given and could not interpret the default waveform.
        TypeError
            - If a parameter's type is incorrect.
        """
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # duration
        if duration is ...:
            duration = self.duration
        else:
            duration = validate_duration(duration)
        
        # io
        if io is ...:
            io = self.io
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # waveform
        if waveform is ...:
            waveform = self.waveform
        else:
            waveform = validate_waveform(waveform)
        
        # Post validate waveform.
        if waveform is None:
            if name.endswith('.ogg'):
                waveform = WAVEFORM_OGG_DEFAULT
            
            else:
                raise ValueError(
                    f'Could not interpret default `waveform` from `name` ({name!r}), please pass it manually.'
                )
        
        # Construct
        new = object.__new__(type(self))
        new.description = description
        new.duration = duration
        new.io = io
        new.name = name
        new.waveform = waveform
        return new
    
    
    def to_data(self):
        """
        Converts the voice attachment to json serializable object.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {
            'id': '0',
        }
        
        put_duration(self.duration, data, False)
        put_description(self.description, data, False)
        put_waveform(self.waveform, data, False)
        return data



__all__ = ()


def process_precreate_parameters_and_raise_extra(keyword_parameters, fields):
    """
    Top level function over ``process_precreate_parameters``.
    
    Parameters
    ----------
    keyword_parameters : `dict<str, object>`
        Keyword parameters passed to a precreate method.
    fields : `dict` of (`str`, `tuple` (`str`, `callable`) items
        Accepted fields by their key as parameters name and their values as `attribute-name - validator`.
    
    Returns
    -------
    processed : `list` of `tuple` (`str`, `object`) items
        The processed values in `attribute-name - value` relation.
    
    Raises
    ------
    BaseException
        - Any exception raised by a field validator.
        - Extra parameters.
    """
    processed = []
    extra = process_precreate_parameters(keyword_parameters, fields, processed)
    raise_extra(extra)
    return processed


def process_precreate_parameters(keyword_parameters, fields, processed):
    """
    Processes keyword parameters passed to a precreate function.
    
    Parameters
    ----------
    keyword_parameters : `dict<str, object>`
        Keyword parameters passed to a precreate method.
    fields : `dict` of (`str`, `tuple` (`str`, `callable`) items
        Accepted fields by their key as parameters name and their values as `attribute-name - validator`.
    processed : `list` of `tuple` (`str`, `object`) items
        Already processed values in `attribute-name - value` relation.
    
    Returns
    -------
    extra : `None`, `dict<str, object>`
        Extra keyword parameters.
    
    Raises
    ------
    BaseException
        - Any exception raised by a field validator.
    """
    extra = None
    
    while keyword_parameters:
        field_name, field_value = keyword_parameters.popitem() 
        try:
            attribute_name, validator = fields[field_name]
        except KeyError:
            if extra is None:
                extra = {}
            extra[field_name] = field_value
            continue
        
        attribute_value = validator(field_value)
        processed.append((attribute_name, attribute_value))
        continue
    
    return extra


def raise_extra(extra):
    """
    Raises exception if there are extra fields.
    
    Parameters
    ----------
    extra : `None`, `dict<str, object>`
        Extra keyword parameters.
    
    Raises
    ------
    TypeError
        - Extra parameters.
    """
    if (extra is not None):
        raise TypeError(
            f'Extra fields received: {extra!r}.'
        )



__all__ = ()

from datetime import datetime as DateTime, timezone as TimeZone

from scarletio import Compound

from ...application import Application
from ...application.application.utils import APPLICATION_FIELD_CONVERTERS
from ...channel import Channel
from ...http import DiscordApiClient
from ...oauth2 import Connection
from ...payload_building import add_payload_fields_from_keyword_parameters, build_edit_payload
from ...user.guild_profile.utils import GUILD_PROFILE_SELF_FIELD_CONVERTERS
from ...user.user.utils import USER_SELF_FIELD_CONVERTERS
from ...utils import datetime_to_timestamp

from ..request_helpers import get_guild_id, get_channel_guild_id_and_id


class ClientCompoundClientEndpoints(Compound):
    
    application : Application
    api : DiscordApiClient
    
    async def edit(self, **keyword_parameters):
        """
        Edits the client. Only the provided parameters will be changed. Every parameter what refers to a user
        account is not tested.
        
        This method is a coroutine.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Additional keyword parameters representing which field of the client should be edited.
        
        Other Parameters
        ----------------
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            An `'jpg'`, `'png'`, `'webp'` image's raw data. If the client is premium account, then it can be
            `'gif'` as well. By passing `None` you can remove the client's current avatar.
        
        avatar_decoration : ``None | AvatarDecoration``, Optional (Keyword only)
            The client's new avatar decoration.
        
        banner : `None`, `bytes-like`, Optional (Keyword only)
            An `'jpg'`, `'png'`, `'webp'`, 'gif'` image's raw data. By passing `None` you can remove the client's
            current avatar.
        
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The new banner color of the client. By passing it as `None` you can remove the client's current one.
        
        display_name : `None`, `str`, Optional (Keyword only)
            The client's non-unique display name.
        
        name : `str`, Optional (Keyword only)
            The client's new name.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        The method's endpoint has long rate limit reset, so consider using timeout and checking rate limits with
        ``RateLimitProxy``.
        """
        data = {}
        add_payload_fields_from_keyword_parameters(USER_SELF_FIELD_CONVERTERS, keyword_parameters, data, True)
        
        if data:
            await self.api.client_edit(data)
    
    
    async def guild_profile_edit(self, guild, *, reason = None, **keyword_parameters):
        """
        Edits the client guild profile in the given guild. Nick and guild specific avatars can be edited on this way.
        
        An extra `reason` is accepted as well, which will show up at the respective guild's audit logs.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``None | int | Guild``
            The guild where the client's nickname will be changed. If `guild` is given as `None`, then the function
            returns instantly.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters representing which field of the guild profile should be edited.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Other Parameters
        ----------------
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            The client's new guild specific avatar.
            
            Can be a `'jpg'`, `'png'`, `'webp'` image's raw data. If the client is premium account, then it can be
            `'gif'` as well. By passing `None` you can remove the client's current avatar.
        
        avatar_decoration : ``None | AvatarDecoration``, Optional (Keyword only)
            The client's new avatar decoration.
        
        banner : `None`, `bytes-like`, Optional (Keyword only)
            The client's new guild specific banner.
            
            Can be a `'jpg'`, `'png'`, `'webp'` and `'gif'`image's raw data.
            By passing `None` you can remove the client's current banner.
        
        bio : `None | str`, Optional (Keyword only)
            The client's new guild specific bio.
        
        nick : `None`, `str`, Optional (Keyword only)
            The client's new nickname. Pass it as `None` to remove it. Empty strings are interpreted as `None`.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = {}
        add_payload_fields_from_keyword_parameters(GUILD_PROFILE_SELF_FIELD_CONVERTERS, keyword_parameters, data, True)
        
        if data:
            await self.api.client_guild_profile_edit(guild_id, data, reason)
    
    
    async def connection_get_all(self):
        """
        Requests the client's connections.
        
        This method is a coroutine.
        
        Returns
        -------
        connections : `list` of ``Connection`` objects
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        For a bot account this request will always return an empty list.
        """
        data = await self.api.client_connection_get_all()
        return [Connection.from_data(connection_data) for connection_data in data]
    
    
    async def guild_leave(self, guild):
        """
        The client leaves the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild from where the client will leave.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        await self.api.guild_leave(guild_id)


    async def join_speakers(self, channel, *, request = False):
        """
        Request to speak or joins the client as a speaker inside of a stage channel. The client must be in the stage
        channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``
            The stage channel to join.
        request : `bool` = `False`, Optional (Keyword only)
            Whether the client should only request to speak.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor as `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        
        if request:
            timestamp = datetime_to_timestamp(DateTime.now(TimeZone.utc))
        else:
            timestamp = None
        
        data = {
            'suppress': False,
            'request_to_speak_timestamp': timestamp,
            'channel_id': channel_id
        }
        
        await self.api.voice_state_edit_own(guild_id, data)
    
    
    async def join_audience(self, channel):
        """
        Moves the client to the audience inside of the stage channel. The client must be in the stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `tuple` (`int`, `int`)
            The stage channel to join.
        
        Raises
        ------
        RuntimeError
            If `channel` is partial.
        TypeError
            If `channel` was not given neither as ``Channel`` nor as `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        
        data = {
            'suppress': True,
            'channel_id': channel_id
        }
        
        await self.api.voice_state_edit_own(guild_id, data)
    
    
    async def application_edit_own(self, application_template = None, **keyword_parameters):
        """
        Edits the client's application.
        
        Parameters
        ----------
        application : ``None | Application`` = `None`, Optional
            Application entity to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to edit the application with.
        
        Other Parameters
        ----------------
        cover : `None`, `bytes-like`, Optional (Keyword only)
            The application's cover.
        
        custom_install_url : `None`, `str`, Optional (Keyword only)
            The application's default custom authorization link if enabled.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        flags : `int`, ``ApplicationFlag``, Optional (Keyword only)
            The application's public flags.
        
        icon : `None`,  `bytes-like`, Optional (Keyword only)
            The application's icon.
        
        install_parameters : `None`, ``ApplicationInstallParameters``, Optional (Keyword only)
            Settings for the application's default in-app authorization link, if enabled.
        
        interaction_endpoint_url : `None`, `str`, Optional (Keyword only)
            Whether and to which url should interaction events be sent to.
        
        role_connection_verification_url : `None`, `str`, Optional (Keyword only)
            The application's role connection verification entry point
        
        tags : `None`, `iterable` of `str`, Optional (Keyword only)
            Up to 5 tags describing the content and functionality of the application.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application = self.application
            
        data = build_edit_payload(
            (None if application.partial else application),
            application_template,
            APPLICATION_FIELD_CONVERTERS,
            keyword_parameters,
        )
        
        if not data:
            return
        
        data = await self.api.application_edit_own(data)
        self.application = application.from_data_own(data)



__all__ = ('ComponentMetadataLabel', )

from scarletio import copy_docs

from .base import ComponentMetadataBase
from .fields import (
    parse_component__label, parse_description, parse_label, put_component__label, put_description, put_label,
    validate_component__label, validate_description, validate_label__label
)


class ComponentMetadataLabel(ComponentMetadataBase):
    """
    Label component metadata.
    
    Attributes
    ----------
    component : ``None | tuple<Component>``
        The contained component.
    
    description : `None | str`
        The description of the component.
    
    label : `None | str`
        The label of the component.
    """
    __slots__ = ('component', 'description', 'label')
    
    
    def __new__(cls, *, component = ..., description = ..., label = ...):
        """
        Creates a new label component metadata.
        
        Parameters
        ----------
        component : ``None | Component``, Optional (Keyword only)
            The contained component.
        
        description : `None | str`, Optional (Keyword only)
            The description of the component.
        
        label : `None | str`, Optional (Keyword only)
            The label of the component.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # component
        if component is ...:
            component = None
        else:
            component = validate_component__label(component)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # label
        if label is ...:
            label = None
        else:
            label = validate_label__label(label)
        
        # Construct
        self = object.__new__(cls)
        self.component = component
        self.description = description
        self.label = label
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            component = keyword_parameters.pop('component', ...),
            description = keyword_parameters.pop('description', ...),
            label = keyword_parameters.pop('label', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # component
        component = self.component
        if (component is not None):
            repr_parts.append(' component = ')
            repr_parts.append(repr(component))
            
            field_added = True
        else:
            field_added = False
        
        # description
        description = self.description
        if (description is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' description = ')
            repr_parts.append(repr(description))
        
        # label
        label = self.label
        if (label is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' label = ')
            repr_parts.append(repr(label))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # component
        component = self.component
        if (component is not None):
            hash_value ^= 1 << 0
            hash_value ^= hash(component)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= 1 << 1
            hash_value ^= hash(description)
        
        # label
        label = self.label
        if (label is not None):
            hash_value ^= 1 << 2
            hash_value ^= hash(label)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # component
        if self.component != other.component:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # label
        if self.label != other.label:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        self.component = parse_component__label(data)
        self.description = parse_description(data)
        self.label = parse_label(data)
        
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        
        put_component__label(self.component, data, defaults, include_internals = include_internals)
        put_description(self.description, data, defaults)
        put_label(self.label, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        
        # component
        component = self.component
        if (component is not None):
            component = component.clean_copy(guild)
        
        new.component = component
        
        # description
        new.description = self.description
        
        # label
        new.label = self.label
        
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # component
        component = self.component
        if (component is not None):
            component = component.copy()
        
        new.component = component
        
        # description
        new.description = self.description
        
        # label
        new.label = self.label
        
        return new
    
    
    def copy_with(self, *, component = ..., description = ..., label = ...):
        """
        Copies the label component metadata with the given fields.
        
        Parameters
        ----------
        component : ``None | Component``, Optional (Keyword only)
            The contained component.
        
        description : `None | str`, Optional (Keyword only)
            The description of the component.
        
        label : `None | str`, Optional (Keyword only)
            The label of the component.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # component
        if component is ...:
            component = self.component
            if (component is not None):
                component = component.copy()
        else:
            component = validate_component__label(component)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # label
        if label is ...:
            label = self.label
        else:
            label = validate_label__label(label)
        
        # Construct
        new = object.__new__(type(self))
        new.component = component
        new.description = description
        new.label = label
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            component = keyword_parameters.pop('component', ...),
            description = keyword_parameters.pop('description', ...),
            label = keyword_parameters.pop('label', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.iter_contents)
    def iter_contents(self):
        component = self.component
        if (component is not None):
            yield from component.iter_contents()




__all__ = ('PollResult',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_answer_id, parse_count, put_answer_id, put_count, validate_answer_id, validate_count, validate_users
)


class PollResult(RichAttributeErrorBaseType):
    """
    Represents a poll's result for an answer.
    
    Attributes
    ----------
    answer_id : `int`
        The represented answer's identifier.
    
    count : `int`
        The amount of votes.
    
    users : ``None | set<ClientUserBase>``
        The known voters.
    """
    __slots__ = ('answer_id', 'count', 'users')
    
    def __new__(cls, *, answer_id = ..., count = ..., users = ...,):
        """
        Creates a poll result instance.
        
        Parameters
        ----------
        answer_id : `int`, Optional (Keyword only)
            The represented answer's identifier.
        
        count : `int`, Optional (Keyword only)
            The amount of votes.
        
        users : `None`, `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The known voters.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # answer_id
        if answer_id is ...:
            answer_id = 0
        else:
            answer_id = validate_answer_id(answer_id)
        
        # count
        if count is ...:
            count = 0
        else:
            count = validate_count(count)
        
        # users
        if users is ...:
            users = None
        else:
            users = validate_users(users)
        
        # post-checks
        if (users is not None) and (len(users) > count):
            count = len(users)
        
        # Construct
        self = object.__new__(cls)
        self.answer_id = answer_id
        self.count = count
        self.users = users
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new poll result with the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Executable data.
        """
        self = object.__new__(cls)
        self.answer_id = parse_answer_id(data)
        self.count = parse_count(data)
        self.users = None
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the poll result to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_answer_id(self.answer_id, data, defaults)
        put_count(self.count, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the count's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # answer_od
        repr_parts.append(' answer_id = ')
        repr_parts.append(repr(self.answer_id))
        
        # count
        repr_parts.append(', count = ')
        repr_parts.append(repr(self.count))
        
        # users
        repr_parts.append(', users = ')
        repr_parts.append(repr(self.users))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two poll results are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # answer_id
        if self.answer_id != other.answer_id:
            return False
        
        # count
        if self.count != other.count:
            return False
        
        # users
        length = len(self)
        if length != len(other) or (length != 0 and self.users != other.users):
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the count's hash."""
        hash_value = 0
        
        # answer_id
        hash_value ^= self.answer_id
        
        # count
        hash_value ^= self.count
        
        # users
        users = self.users
        if (users is not None):
            hash_value ^= len(users) << 4
            for user in users:
                hash_value ^= hash(user)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the poll result.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        new.answer_id = self.answer_id
        new.count = self.count
        
        users = self.users
        if (users is not None):
            users = users.copy()
        new.users = users
        
        return new
    
    
    def copy_with(self, *, answer_id = ..., count = ..., users = ...):
        """
        Copies the poll result with the given fields.
        
        Parameters
        ----------
        answer_id : `int`, Optional (Keyword only)
            The represented answer's identifier.
        
        count : `int`, Optional (Keyword only)
            The amount of votes.
        
        users : `None`, `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The known voters.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # answer_id
        if answer_id is ...:
            answer_id = self.answer_id
        else:
            answer_id = validate_answer_id(answer_id)
        
        # count
        if count is ...:
            count = self.count
        else:
            count = validate_count(count)
        
        # users
        if users is ...:
            users = self.users
            if (users is not None):
                users = users.copy()
        else:
            users = validate_users(users)
        
        # post-checks
        if (users is not None) and len(users) > count:
            count = len(users)
        
        # Construct
        new = object.__new__(type(self))
        new.answer_id = answer_id
        new.count = count
        new.users = users
        return new
    
    
    def __len__(self):
        """Returns the amount of known voters."""
        users = self.users
        if users is None:
            return 0
        
        return len(users)
    
    
    def __iter__(self):
        """Iterates over the known voters."""
        users = self.users
        if (users is not None):
            yield from users
    
    
    def __contains__(self, user):
        """Returns whether the given user is in the known votes."""
        users = self.users
        if (users is None):
            return False
        
        return user in users
    
    
    def _merge_with(self, other):
        """
        Merges the poll result with the other one.
        
        Always `other` is right and `other` should be empty.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            Other instance.
        """
        new_count = other.count
        
        self_users = self.users
        other_users = other.users
        if self_users is None:
            if (other_users is not None):
                self_users = other_users.copy()
                self.users = self_users
        
        else:
            if self.count != new_count:
                self_users.clear()
                if (other_users is not None):
                    self_users.update(other_users)
            
            else:
                if (other_users is not None) and (not other_users.issubset(self_users)):
                    self_users.clear()
                    self_users.update(other_users)
            
        self.count = new_count
    
    
    @classmethod
    def _create_empty(cls, answer_id):
        """
        Creates an empty poll result.
        
        Parameters
        ----------
        answer_id : `int`
            The represented answer's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.answer_id = answer_id
        self.count = 0
        self.users = None
        return self
    
    
    def _add_vote(self, user):
        """
        Adds a voter.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user who voted.
        
        Returns
        -------
        success : `bool`
        """
        users = self.users
        if users is None:
            self.users = {user}
            self.count += 1
            return True
        
        old_length = len(users)
        users.add(user)
        new_length = len(users)
        if old_length == new_length:
            return False
        
        self.count += 1
        return True
    
    
    def _fill_some_votes(self, some_users):
        """
        Fills out some voters.
        
        Parameters
        ----------
        some_users : ``list<ClientUserBase>``
            The users who voted.
        """
        users = self.users
        if users is None:
            if some_users:
                users = {*some_users}
                self.users = users
        else:
            users.update(some_users)
        
        # This should not happen
        if (users is not None) and (self.count < len(users)):
            self.count = len(users)
    
    
    def _fill_all_votes(self, all_users):
        """
        Fills out all voters.
        
        Parameters
        ----------
        all_users : ``list<ClientUserBase>``
            The users who voted.
        """
        users = self.users
        if users is None:
            if all_users:
                users = {*all_users}
                self.users = users
        else:
            users.clear()
            users.update(all_users)
        
        # This should not happen
        user_count = 0 if users is None else len(users)
        if self.count != user_count:
            self.count = user_count
    
    
    def _remove_vote(self, user):
        """
        Removes a voter.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user who removed their vote.
        
        Returns
        -------
        success : `bool`
        """
        users = self.users
        count = self.count
        
        if users is None:
            if count <= 0:
                return False
            
            self.count = count - 1
            return True
        
        old_length = len(users)
        users.discard(user)
        new_length = len(users)
        if old_length == new_length:
            count = self.count
            if count <= new_length:
                return False
            
            self.count = count - 1
            return True
        
        if count <= 0:
            return False
        
        self.count = count - 1
        return True
    
    
    @property
    def unknown(self):
        """
        Returns the amount of unknown voters.
        
        Returns
        -------
        unknown : `int`
        """
        count = self.count
        users = self.users
        if users is None:
            return count
        
        return count - len(users)
    
    
    def filter_after(self, limit, after):
        """
        If we know all the voters, then instead of executing a Discord API request we filter the voters locally
        using this method.
        
        Parameters
        ----------
        limit : `int`
            The maximal limit of the users to return.
        after : `int`
            Gets the users after this specified id.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        """
        users = self.users
        if users is None:
            return []
        
        filtered_users = [user for user in self if user.id > after]
        filtered_users.sort()
        del filtered_users[limit:]
        return filtered_users
    
    
    def iter_users(self):
        """
        Iterates over the known voters.
        
        This method is an iterable generator.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        users = self.users
        if (users is not None):
            yield from users



__all__ = ('InteractionComponentMetadataMentionableSelect',)

from scarletio import copy_docs, include

from .string_select import InteractionComponentMetadataStringSelect


ComponentType = include('ComponentType')


class InteractionComponentMetadataMentionableSelect(InteractionComponentMetadataStringSelect):
    """
    Interaction component metadata representing a mentionable (role and user) select component.
    
    Attributes
    ----------
    custom_id : `None | str`
        Custom identifier to detect which component was clicked (or used) by the user.
    
    values : `None | tuple<str>`
        The component's values defined by the user.
    """
    __slots__ = ()

    @copy_docs(InteractionComponentMetadataStringSelect.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        custom_id = self.custom_id
        if (custom_id is not None):
            yield (custom_id, ComponentType.mentionable_select, self.values)



__all__ = ()

from functools import partial as partial_func

from .fields import (
    put_activity_application_ids, put_banner_color, put_description, put_name, put_privacy_level, put_tags,
    validate_activity_application_ids, validate_banner_color, validate_description, validate_name,
    validate_privacy_level, validate_tags
)
from .guild_activity_overview import GUILD_ACTIVITY_OVERVIEW_DISCOVERY_SPLASH, GUILD_ACTIVITY_OVERVIEW_ICON


GUILD_ACTIVITY_OVERVIEW_FIELD_CONVERTERS = {
    'activity_application_ids': (validate_activity_application_ids, put_activity_application_ids),
    'banner_color': (validate_banner_color, put_banner_color),
    'description': (validate_description, put_description),
    'discovery_splash': (
        partial_func(GUILD_ACTIVITY_OVERVIEW_DISCOVERY_SPLASH.validate_icon, allow_data = True),
        partial_func(GUILD_ACTIVITY_OVERVIEW_DISCOVERY_SPLASH.put_into, as_data = True),
    ),
    'icon': (
        partial_func(GUILD_ACTIVITY_OVERVIEW_ICON.validate_icon, allow_data = True),
        partial_func(GUILD_ACTIVITY_OVERVIEW_ICON.put_into, as_data = True),
    ),
    'name': (validate_name, put_name),
    'privacy_level': (validate_privacy_level, put_privacy_level),
    'tags': (validate_tags, put_tags),
}
